# coding=utf-8
# Copyright 2021 The Trax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Tests for RL training."""

import functools
import math
import os
import pickle

from absl.testing import absltest
import tensorflow as tf

from trax import layers as tl
from trax import models
from trax import optimizers as opt
from trax import test_utils
from trax.rl import task as rl_task
from trax.rl import training
from trax.supervised import lr_schedules


class TrainingTest(absltest.TestCase):

  def setUp(self):
    super().setUp()
    test_utils.ensure_flag('test_tmpdir')
    self._model_fn = functools.partial(
        models.Policy,
        body=lambda mode: tl.Serial(  # pylint: disable=g-long-lambda
            tl.Dense(64), tl.Relu(), tl.Dense(64), tl.Relu()
        ),
    )

  def test_policy_gradient_smoke(self):
    """Smoke test of PolicyGradient."""
    task = rl_task.RLTask('CartPole-v0', max_steps=2)
    tmp_dir = self.create_tempdir().full_path
    agent = training.PolicyGradient(
        task,
        model_fn=self._model_fn,
        optimizer=opt.Adam,
        batch_size=2,
        n_trajectories_per_epoch=2,
        n_eval_episodes=1,
        output_dir=tmp_dir)
    agent.run(1)
    self.assertEqual(agent.current_epoch, 1)

  def test_expert_iteration_smoke(self):
    """Smoke test of ExpertIteration."""
    task = rl_task.RLTask('CartPole-v0', max_steps=2)
    tmp_dir = self.create_tempdir().full_path
    agent = training.ExpertIteration(
        task,
        model_fn=self._model_fn,
        optimizer=opt.Adam,
        batch_size=2,
        n_trajectories_per_epoch=2,
        n_train_steps_per_epoch=2,
        n_eval_episodes=1,
        output_dir=tmp_dir,
    )
    agent.run(1)
    self.assertEqual(agent.current_epoch, 1)

  def test_policy_gradient_save_restore(self):
    """Check save and restore of policy agent."""
    task = rl_task.RLTask('CartPole-v0', max_steps=2)
    tmp_dir = self.create_tempdir().full_path
    agent1 = training.PolicyGradient(
        task,
        model_fn=self._model_fn,
        optimizer=opt.Adam,
        batch_size=2,
        n_trajectories_per_epoch=2,
        n_eval_episodes=1,
        output_dir=tmp_dir)
    agent1.run(1)
    agent1.run(1)
    self.assertEqual(agent1.current_epoch, 2)
    self.assertEqual(agent1.loop.step, 2)
    # Trainer 2 starts where agent 1 stopped.
    agent2 = training.PolicyGradient(
        task,
        model_fn=self._model_fn,
        optimizer=opt.Adam,
        batch_size=2,
        n_trajectories_per_epoch=2,
        n_eval_episodes=1,
        output_dir=tmp_dir)
    agent2.run(1)
    self.assertEqual(agent2.current_epoch, 3)
    self.assertEqual(agent2.loop.step, 3)
    # Manually set saved epoch to 1.
    dictionary = {
        'epoch': 1,
        'avg_returns': [0.0],
        'avg_returns_temperature_0.0': {
            200: [0.0]
        }
    }
    with tf.io.gfile.GFile(os.path.join(tmp_dir, 'rl.pkl'), 'wb') as f:
      pickle.dump(dictionary, f)
    # Trainer 3 restores from a checkpoint with Agent/Loop step mistmatch,
    # should fail.
    def agent3_fn():
      return training.PolicyGradient(
          task,
          model_fn=self._model_fn,
          optimizer=opt.Adam,
          batch_size=2,
          n_trajectories_per_epoch=2,
          n_eval_episodes=1,
          output_dir=tmp_dir,
      )
    self.assertRaises(ValueError, agent3_fn)
    agent1.close()
    agent2.close()

  def test_policy_gradient_cartpole(self):
    """Trains a policy on cartpole."""
    task = rl_task.RLTask('CartPole-v0', max_steps=200)
    lr = lambda: lr_schedules.multifactor(constant=1e-2, factors='constant')
    max_avg_returns = -math.inf
    for _ in range(2):
      agent = training.PolicyGradient(
          task,
          model_fn=self._model_fn,
          optimizer=opt.Adam,
          lr_schedule=lr,
          batch_size=128,
          eval_temperatures=[0.0, 0.5],
          n_eval_episodes=1,
          n_trajectories_per_epoch=2,
      )
      # Assert that we get to 200 at some point and then exit so the test is as
      # fast as possible.
      for ep in range(200):
        agent.run(1)
        self.assertEqual(agent.current_epoch, ep + 1)
        if agent.avg_returns[-1] == 200.0:
          for eval_t in agent._eval_temperatures:
            self.assertEqual(
                len(agent._avg_returns_temperatures[eval_t][200]),
                len(agent.avg_returns))
          return
      max_avg_returns = max(max_avg_returns, agent.avg_returns[-1])
    self.fail(
        'The expected score of 200 has not been reached. '
        'Maximum at end was {}.'.format(max_avg_returns)
    )

  def test_dqntrainer_cartpole(self):
    """Test-runs joint PPO on CartPole."""

    task = rl_task.RLTask('CartPole-v0', initial_trajectories=0,
                          max_steps=2)
    value_body = lambda mode: tl.Serial(tl.Dense(64), tl.Relu())

    lr = lambda: lr_schedules.multifactor(  # pylint: disable=g-long-lambda
        constant=1e-2, warmup_steps=100, factors='constant * linear_warmup')

    trainer = training.DQN(
        task,
        value_body=value_body,
        value_optimizer=opt.Adam,
        value_lr_schedule=lr,
        value_batch_size=4,
        value_train_steps_per_epoch=2,
        n_trajectories_per_epoch=5)
    trainer.run(2)
    self.assertEqual(2, trainer.current_epoch)


if __name__ == '__main__':
  absltest.main()



# coding=utf-8
# Copyright 2021 The Trax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Trax trainer."""
import atexit
import datetime
import functools
import os

from absl import app
from absl import flags
from absl import logging

import gin
import jax
from jax.lib import xla_extension as xc
import tensorflow.compat.v2 as tf
from trax import fastmath
from trax import trainer_flags  # pylint: disable=unused-import
from trax.supervised import trainer_lib
from trax.tf_numpy import numpy as tf_np

FLAGS = flags.FLAGS
Backend = fastmath.Backend


# TODO(afrozm): Share between trainer.py and rl_trainer.py
def _tf_setup_from_flags():
  """Processes TensorFlow-relevant flags."""
  if FLAGS.enable_eager_execution:
    tf.compat.v1.enable_eager_execution()
  if FLAGS.tf_xla:
    tf.config.optimizer.set_jit(True)
    fastmath.tf.set_tf_xla_forced_compile(FLAGS.tf_xla_forced_compile)
  tf.config.optimizer.set_experimental_options({
      'pin_to_host_optimization': FLAGS.tf_opt_pin_to_host,
      'layout_optimizer': FLAGS.tf_opt_layout,
  })
  tf_np.set_allow_float64(FLAGS.tf_allow_float64)


# TODO(afrozm): Share between trainer.py and rl_trainer.py
def _gin_parse_configs():
  """Initializes gin-controlled bindings."""
  # Imports for configurables
  # pylint: disable=g-import-not-at-top,unused-import,g-bad-import-order,reimported,unused-variable
  from trax import models as _trax_models
  from trax import optimizers as _trax_opt
  # pylint: disable=g-import-not-at-top,unused-import,g-bad-import-order,reimported,unused-variable

  configs = FLAGS.config if FLAGS.config is not None else []
  # Override with --dataset and --model
  if FLAGS.dataset:
    configs.append("data_streams.dataset_name='%s'" % FLAGS.dataset)
  if FLAGS.data_dir:
    configs.append("data_streams.data_dir='%s'" % FLAGS.data_dir)
  if FLAGS.model:
    configs.append('train.model=@trax.models.%s' % FLAGS.model)
  gin.parse_config_files_and_bindings(FLAGS.config_file, configs)


def _output_dir_or_default():
  """Returns a path to the output directory."""
  if FLAGS.output_dir:
    output_dir = FLAGS.output_dir
    trainer_lib.log('Using --output_dir {}'.format(output_dir))
    return os.path.expanduser(output_dir)

  # Else, generate a default output dir (under the user's home directory).
  try:
    dataset_name = gin.query_parameter('data_streams.dataset_name')
  except ValueError:
    dataset_name = 'random'
  output_name = '{model_name}_{dataset_name}_{timestamp}'.format(
      model_name=gin.query_parameter('train.model').configurable.name,
      dataset_name=dataset_name,
      timestamp=datetime.datetime.now().strftime('%Y%m%d_%H%M'),
  )
  output_dir = os.path.join('~', 'trax', output_name)
  output_dir = os.path.expanduser(output_dir)
  print()
  trainer_lib.log('No --output_dir specified')
  trainer_lib.log('Using default output_dir: {}'.format(output_dir))
  return output_dir


# TODO(afrozm): Share between trainer.py and rl_trainer.py
def _jax_and_tf_configure_for_devices():  # pylint: disable=missing-function-docstring
  if FLAGS.use_tpu:
    jax.config.update('jax_platform_name', 'tpu')
    jax.config.update('jax_xla_backend', FLAGS.jax_xla_backend)
    jax.config.update('jax_backend_target', FLAGS.jax_backend_target)
  if (FLAGS.enable_eager_execution and (fastmath.is_backend(Backend.NUMPY) or
                                        fastmath.is_backend(Backend.JAX))):
    # Numpy backend doesn't benefit from having the input pipeline run on GPU,
    # and jax backend has GPU memory contention if TF uses the GPU. Gin must be
    # set up first before determining the backend.
    tf.config.experimental.set_visible_devices([], 'GPU')


def _train_using_tf(output_dir):
  worker_cpu = tf_init_tpu()
  with tf.device(worker_cpu):
    if trainer_lib.num_devices() == 1:
      # TF's device priority is GPU > CPU > TPU, so we need to explicitly make
      # the TPU core the default device here.
      with tf.device('/device:TPU:0'):
        trainer_lib.train(output_dir=output_dir)
    else:
      trainer_lib.train(output_dir=output_dir)


@gin.configurable
def tf_init_tpu(worker='', protocol=None):
  """Initializes TPU for TensorFlow.

  Args:
    worker: The BNS address of the remote TPU worker. If it's empty (the default
      value), TF will assume the TPU devices are connected to the local host.
    protocol: The network protocol used to connect to the TPU worker.
  Returns:
    The device name of the TPU worker's CPU.
  """
  protocol = protocol or 'grpc'
  is_local = (worker in ('', 'local'))
  resolver = tf.distribute.cluster_resolver.TPUClusterResolver(tpu=worker)
  if not is_local:
    tf.config.experimental_connect_to_cluster(resolver, protocol=protocol)
  tf.tpu.experimental.initialize_tpu_system(resolver)
  if is_local:
    return ''
  else:
    return '/job:worker'


def _make_jax_gpu_cluster(host_id, server_ip, n_hosts, server_port=5005):
  """Make JAX GPU Cluster."""

  addr = f'{server_ip}:{server_port}'
  if host_id == 0:
    logging.info('starting service on %s', addr)
    service = xc.get_distributed_runtime_service(addr, n_hosts)
    # We add an explicit call to shutdown the service via atexit as Python
    # interpreter may not call the service destructor on process termination.
    atexit.register(service.shutdown)

  logging.info('connecting to service on %s', addr)
  dist_client = xc.get_distributed_runtime_client(addr, host_id)
  dist_client.connect()
  atexit.register(dist_client.shutdown)

  # register dist gpu backend
  factory = functools.partial(jax.lib.xla_client.make_gpu_client,
                              dist_client, host_id)
  jax.lib.xla_bridge.register_backend_factory('gpu', factory, priority=300)


def main(_):
  logging.set_verbosity(FLAGS.log_level)

  _tf_setup_from_flags()
  _gin_parse_configs()
  _jax_and_tf_configure_for_devices()

  # Create a JAX GPU cluster if using JAX and given a chief IP.
  if fastmath.is_backend(Backend.JAX) and FLAGS.gpu_cluster_chief_ip:
    _make_jax_gpu_cluster(FLAGS.gpu_cluster_host_id,
                          FLAGS.gpu_cluster_chief_ip,
                          FLAGS.gpu_cluster_n_hosts,
                          FLAGS.gpu_cluster_port)

  if FLAGS.disable_jit:
    fastmath.disable_jit()

  output_dir = _output_dir_or_default()
  if FLAGS.use_tpu and fastmath.is_backend(Backend.TFNP):
    _train_using_tf(output_dir)
  else:
    trainer_lib.train(output_dir=output_dir)

  trainer_lib.log('Finished training.')


if __name__ == '__main__':
  app.run(main)



# coding=utf-8
# Copyright 2021 The Trax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Layers: trainable functions as neural network building blocks."""

import gin
# We create a flat layers.* namespace for uniform calling conventions as we
# upstream changes.
# pylint: disable=wildcard-import
from trax.layers.acceleration import *
from trax.layers.activation_fns import *
from trax.layers.assert_shape import *
from trax.layers.attention import *
from trax.layers.base import *
from trax.layers.combinators import *
from trax.layers.convolution import *
from trax.layers.core import *
from trax.layers.deconvolution import *
from trax.layers.initializers import *
from trax.layers.metrics import *
from trax.layers.normalization import *
from trax.layers.pooling import *
from trax.layers.research.efficient_attention import *
from trax.layers.research.position_encodings import *
from trax.layers.research.rel_attention import *
from trax.layers.research.resampling import *
from trax.layers.research.rotary_positional_embedding import *
from trax.layers.research.sparsity import *
from trax.layers.reversible import *
from trax.layers.rnn import *


# Ginify
def layer_configure(*args, **kwargs):
  kwargs['module'] = 'trax.layers'
  return gin.external_configurable(*args, **kwargs)

# pylint: disable=used-before-assignment
# pylint: disable=invalid-name
Relu = layer_configure(Relu)
Gelu = layer_configure(Gelu)
FastGelu = layer_configure(FastGelu)
Sigmoid = layer_configure(Sigmoid)
Tanh = layer_configure(Tanh)
HardSigmoid = layer_configure(HardSigmoid)
HardTanh = layer_configure(HardTanh)
Exp = layer_configure(Exp)
LogSoftmax = layer_configure(LogSoftmax)
Softmax = layer_configure(Softmax)
Softplus = layer_configure(Softplus)
L2Loss = layer_configure(L2Loss)
LSTMCell = layer_configure(LSTMCell)
GRUCell = layer_configure(GRUCell)

BatchNorm = layer_configure(BatchNorm)
LayerNorm = layer_configure(LayerNorm)
FilterResponseNorm = layer_configure(FilterResponseNorm)
ThresholdedLinearUnit = layer_configure(ThresholdedLinearUnit)

Attention = layer_configure(Attention, denylist=['mode'])
CausalAttention = layer_configure(CausalAttention, denylist=['mode'])
FavorAttention = layer_configure(FavorAttention, denylist=['mode'])
Favor = layer_configure(Favor, denylist=['mode'])
CausalFavor = layer_configure(CausalFavor, denylist=['mode'])
CausalFavorAttention = layer_configure(CausalFavorAttention, denylist=['mode'])
DotProductCausalAttention = layer_configure(
    DotProductCausalAttention, denylist=['mode'])
SelfAttention = layer_configure(SelfAttention, denylist=['mode'])
ModularCausalAttention = layer_configure(ModularCausalAttention,
                                         denylist=['mode'])
LowRankCausalAttention = layer_configure(LowRankCausalAttention,
                                         denylist=['mode'])
MultiplicativeCausalAttention = layer_configure(MultiplicativeCausalAttention,
                                                denylist=['mode'])
MultiplicativeModularCausalAttention = layer_configure(
    MultiplicativeModularCausalAttention, denylist=['mode'])
ConvCausalAttention = layer_configure(ConvCausalAttention, denylist=['mode'])
MultiplicativeConvCausalAttention = layer_configure(
    MultiplicativeConvCausalAttention, denylist=['mode'])
ConvTranspose = layer_configure(ConvTranspose)
LSHSelfAttention = layer_configure(LSHSelfAttention, denylist=['mode'])
PureLSHSelfAttention = layer_configure(PureLSHSelfAttention, denylist=['mode'])
MixedLSHSelfAttention = layer_configure(
    MixedLSHSelfAttention, denylist=['mode'])
PureLSHSelfAttentionWrapper = layer_configure(
    PureLSHSelfAttentionWrapper, denylist=['mode'])
EncDecAttention = layer_configure(EncDecAttention, denylist=['mode'])

PositionalEncoding = layer_configure(
    PositionalEncoding, denylist=['mode'])
InfinitePositionalEncoding = layer_configure(
    InfinitePositionalEncoding, denylist=['mode'])
TimeBinPositionalEncoding = layer_configure(
    TimeBinPositionalEncoding, denylist=['mode'])

AtariConvInit = layer_configure(AtariConvInit)
CrossEntropyLossWithLogSoftmax = layer_configure(CrossEntropyLossWithLogSoftmax)
WeightedCategoryAccuracy = layer_configure(WeightedCategoryAccuracy)
SequenceAccuracy = layer_configure(SequenceAccuracy)
CategoryCrossEntropy = layer_configure(CategoryCrossEntropy)
WeightedCategoryCrossEntropy = layer_configure(WeightedCategoryCrossEntropy)
MacroAveragedFScore = layer_configure(MacroAveragedFScore)
WeightedFScore = layer_configure(WeightedFScore)
RelativeAttentionLayer = layer_configure(RelativeAttentionLayer)
RelativeAttentionLMLayer = layer_configure(RelativeAttentionLMLayer)
RelativeAttentionWrapper = layer_configure(RelativeAttentionWrapper)
Rotate = layer_configure(Rotate)
AveragePooling = layer_configure(AveragePooling)
LinearPooling = layer_configure(LinearPooling)
LinearUpsampling = layer_configure(LinearUpsampling)
NoUpsampling = layer_configure(NoUpsampling)
NaiveUpsampling = layer_configure(NaiveUpsampling)
AttentionResampling = layer_configure(AttentionResampling)



# coding=utf-8
# Copyright 2021 The Trax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Trax initializers."""

from absl import logging

import numpy as np
import tensorflow.compat.v2 as tf
from trax.fastmath import numpy as jnp
from trax.fastmath import random


def _GetFans(shape, out_dim=-1, in_dim=-2, nonreceptive_dims=None):
  """Get the fan-in and fan-out sizes for the given shape and dims."""
  # Temporary fix until numpy.delete supports negative indices.
  if out_dim < 0:
    out_dim += len(shape)
  if in_dim < 0:
    in_dim += len(shape)

  if nonreceptive_dims is None:
    nonreceptive_dims = []
  if not isinstance(nonreceptive_dims, (list, tuple)):
    nonreceptive_dims = [nonreceptive_dims]

  receptive_field = jnp.prod(np.delete(shape, [in_dim, out_dim,
                                               *nonreceptive_dims]))
  if len(shape) >= 2:
    fan_in, fan_out = shape[in_dim], shape[out_dim]
  elif len(shape) == 1:
    fan_in = shape[0]
    fan_out = shape[0]
  else:
    fan_in = 1.
    fan_out = 1.
    fan_in *= receptive_field
    fan_out *= receptive_field
  return fan_in, fan_out


def InitializerFromFile(path):
  """Loads parameters from .npy file."""

  def Initializer(shape, rng):
    del rng
    logging.info('Loading pretrained embeddings from %s', path)
    with tf.io.gfile.GFile(path, 'rb') as f:
      parameters = jnp.load(f)
    assert jnp.shape(parameters) == shape, (
        'Expected shape %s, got %s' % (shape, jnp.shape(parameters)))
    return parameters

  return Initializer


def _PureShape(shape):
  """Make sure shape does not contain int tensors by calling int()."""
  return [int(x) for x in shape]


def RandomNormalInitializer(stddev=1e-2):
  """Returns an initializer for random normal coefficients."""
  return lambda shape, rng: (stddev * random.normal(  # pylint: disable=g-long-lambda
      rng, _PureShape(shape)).astype('float32'))


def RandomUniformInitializer(lim=1.0):
  """Returns an initializer for random uniform coefficients."""
  # Make sure shape does not contain int tensors by calling int() below.
  return lambda shape, rng: random.uniform(  # pylint: disable=g-long-lambda
      rng, _PureShape(shape), jnp.float32, -lim, lim)


def ScaledInitializer(out_dim, in_dim, scale, mode, distribution):
  """Returns an initializer that adjusts its scale based on weight shapes."""
  if scale <= 0.:
    raise ValueError('scale must be positive float, {} given'.format(scale))
  if mode not in {'fan_in', 'fan_out', 'fan_avg'}:
    raise ValueError(
        'Invalid mode argument:, {}, must be either fan_in, fan_out or fan_avg'
        .format(mode))

  def Init(shape, rng, nonreceptive_dims=None):
    """Returns random values for initializing weights of the given `shape`."""
    shape = _PureShape(shape)
    fan_in, fan_out = _GetFans(shape, out_dim, in_dim, nonreceptive_dims)
    gain = scale
    if mode == 'fan_in':
      gain /= fan_in
    elif mode == 'fan_out':
      gain /= fan_out
    elif mode == 'fan_avg':
      gain /= (fan_in + fan_out) / 2
    if distribution == 'truncated_normal':
      # constant from scipy.stats.truncnorm.std(a=-2, b=2, loc=0., scale=1.)
      stddev = jnp.sqrt(gain) / .87962566103423978
      new_weights = random.truncated_normal(rng, -2, 2, shape) * stddev
      return new_weights.astype('float32')
    elif distribution == 'normal':
      new_weights = random.normal(rng, shape) * jnp.sqrt(gain)
      return new_weights.astype('float32')
    elif distribution == 'uniform':
      lim = jnp.sqrt(3. * gain)
      return random.uniform(rng, shape, jnp.float32, -lim, lim)
    else:
      raise ValueError('invalid distribution for ScaleInitializer')

  return Init


def GlorotNormalInitializer(out_dim=-1, in_dim=-2, scale=1.):
  """Returns an initializer for random Glorot-scaled coefficients."""
  return ScaledInitializer(out_dim, in_dim, scale, 'fan_avg', 'normal')


def GlorotUniformInitializer(out_dim=-1, in_dim=-2, scale=1.):
  """Returns an initializer for random uniform Glorot-scaled coefficients."""
  return ScaledInitializer(out_dim, in_dim, scale, 'fan_avg', 'uniform')


def LeCunNormalInitializer(out_dim=-1, in_dim=-2, scale=1.):
  """Returns an initializer for random LeCun-scaled coefficients."""
  return ScaledInitializer(out_dim, in_dim, scale, 'fan_in', 'normal')


def LeCunUniformInitializer(out_dim=-1, in_dim=-2, scale=1.):
  """Returns an initializer for random uniform LeCun-scaled coefficients."""
  return ScaledInitializer(out_dim, in_dim, scale, 'fan_in', 'uniform')


def KaimingNormalInitializer(out_dim=-1, in_dim=-2, param=0.):
  """Returns an initializer for random Kaiming-scaled coefficients."""
  return ScaledInitializer(
      out_dim, in_dim, 2.0 / jnp.sqrt(1 + param**2), 'fan_in', 'normal')


def KaimingUniformInitializer(out_dim=-1, in_dim=-2, param=0.):
  """Returns an initializer for random uniform Kaiming-scaled coefficients."""
  return ScaledInitializer(
      out_dim, in_dim, 2.0 / jnp.sqrt(1 + param**2), 'fan_in', 'uniform')


def OrthogonalInitializer(stddev=1.0):
  """Returns an orthogonal initializer."""
  def Init(shape, rng):
    """Returns orthogonalized random normal values with the given `shape`."""
    # Have at least 2 elements in shape.
    cur_shape = list(shape)
    while len(cur_shape) < 2:
      cur_shape = [1] + cur_shape

    # Flatten the input shape with the last dimension remaining.
    n_rows = 1
    for dim in cur_shape[:-1]:
      n_rows *= dim
    n_cols = cur_shape[-1]
    flat_shape = (n_cols, n_rows) if n_rows < n_cols else (n_rows, n_cols)

    # Generate a random matrix
    a = random.normal(rng, flat_shape, dtype=jnp.float32)

    # Compute the qr factorization
    q, r = jnp.linalg.qr(a)

    # Make Q uniform
    d = jnp.diag(r)
    q *= jnp.sign(d)

    # Transpose and reshape back q if needed.
    if n_rows < n_cols:
      q = jnp.transpose(q)
    q = jnp.reshape(q, shape)

    # Return scaled as requested.
    return stddev * q

  return Init


def AtariConvInit(kernel_shape, rng, dtype=jnp.float32):
  """The standard init for Conv laters and Atari."""
  filter_height, filter_width, fan_in, _ = kernel_shape
  std = 1 / jnp.sqrt(fan_in * filter_height * filter_width)
  return random.uniform(rng, kernel_shape, dtype, minval=-std, maxval=std)



# coding=utf-8
# Copyright 2021 The Trax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""A number of RL functions intended to be later wrapped as Trax layers.

  Wrapping happens with help of the function tl.Fn.
"""

from trax.fastmath import numpy as jnp


def ValueLoss(values, returns, value_loss_coeff):
  """Definition of the loss of the value function."""
  advantages = returns - values
  l2_value_loss = jnp.mean(advantages**2) * value_loss_coeff
  return l2_value_loss


def ExplainedVariance(values, returns):
  """Definition of explained variance - an approach from OpenAI baselines."""
  assert returns.shape == values.shape, (
      f'returns.shape was {returns.shape} and values.shape was {values.shape}')
  # TODO(henrykm): it would be good to explain the relation with the time dim.
  returns_variance = jnp.var(returns)
  explained_variance = 1 - jnp.var(returns-values)/returns_variance
  return explained_variance


def PreferredMove(dist_inputs, sample):
  """Definition of the preferred move."""
  preferred_moves = sample(dist_inputs, temperature=0.0)
  return jnp.mean(preferred_moves)


def NewLogProbs(dist_inputs, actions, log_prob_fun):
  """Given distribution and actions calculate log probs."""
  new_log_probs = log_prob_fun(dist_inputs,
                               actions)
  return new_log_probs


# TODO(henrykm): Clarify how jnp.mean is applied.
def EntropyLoss(dist_inputs, distribution, coeff):
  """Definition of the Entropy Layer."""
  entropy_loss = distribution.entropy(dist_inputs) * coeff
  return jnp.mean(entropy_loss)


def ProbsRatio(dist_inputs, actions, old_log_probs, log_prob_fun):
  """Probability Ratio from the PPO algorithm."""
  # dist_inputs of the shape float32[128,1,18]
  # actions of the shape int32[128,1]
  # and old_log_probs of the shape float32[128,1]
  new_log_probs = NewLogProbs(dist_inputs, actions, log_prob_fun)
  assert new_log_probs.shape == old_log_probs.shape, (
      f'new_log_probs.shape was {new_log_probs.shape} and'
      f'old_log_probs.shape was {old_log_probs.shape}')
  # The ratio between new_probs and old_probs expressed
  # using log_probs and exponentiation
  probs_ratio = jnp.exp(new_log_probs - old_log_probs)
  return probs_ratio


def ApproximateKLDivergence(dist_inputs, actions, old_log_probs, log_prob_fun):
  """Probability Ratio from the PPO algorithm."""
  new_log_probs = NewLogProbs(dist_inputs, actions, log_prob_fun)
  assert new_log_probs.shape == old_log_probs.shape, (
      f'new_log_probs.shape was {new_log_probs.shape} and'
      f'old_log_probs.shape was {old_log_probs.shape}')
  approximate_kl_divergence = 0.5 * \
      jnp.mean(new_log_probs - old_log_probs) ** 2
  return approximate_kl_divergence


def UnclippedObjective(probs_ratio, advantages):
  """Unclipped Objective from the PPO algorithm."""
  assert probs_ratio.shape == advantages.shape, (
      f'probs_ratio.shape was {probs_ratio.shape} and'
      f'advantages.shape was {advantages.shape}')
  unclipped_objective = probs_ratio * advantages
  return unclipped_objective


def ClippedObjective(probs_ratio, advantages, epsilon):
  """Clipped Objective from the PPO algorithm."""
  assert probs_ratio.shape == advantages.shape, (
      f'probs_ratio.shape was {probs_ratio.shape} and'
      f'advantages.shape was {advantages.shape}')
  clipped_objective = jnp.clip(probs_ratio, 1 - epsilon,
                               1 + epsilon) * advantages
  assert probs_ratio.shape == clipped_objective.shape, (
      f'probs_ratio.shape was {probs_ratio.shape} and'
      f'clipped_objective.shape was {clipped_objective.shape}')
  return clipped_objective


def PPOObjective(dist_inputs, values, returns, dones, rewards,
                 actions, old_log_probs, log_prob_fun, epsilon,
                 normalize_advantages):
  """PPO Objective."""
  # dist_inputs of the shape float32[128,1,18]
  # values of the shape float32[128,1,1]
  # returns of the shape float32[128,1,1]
  # dones of the shape float32[128,1,1]
  # rewards of the shape int32[128,1,1]
  # actions of the shape int32[128,1]
  # and old_log_probs of the shape float32[128,1]
  returns = returns.squeeze(axis=2)
  values = values.squeeze(axis=2)
  dones = dones.squeeze(axis=2)
  rewards = rewards.squeeze(axis=2)
  assert rewards.shape == dones.shape, (
      f'rewards.shape was {rewards.shape} and dones.shape was {dones.shape}')
  assert dones.shape == values.shape, (
      f'dones.shape was {dones.shape} and values.shape was {values.shape}')
  assert returns.shape == values.shape, (
      f'returns.shape was {returns.shape} and values.shape was {values.shape}')
  assert returns.shape == old_log_probs.shape, (
      f'returns.shape was {returns.shape} and'
      f'old_log_probs.shape was {old_log_probs.shape}')

  probs_ratio = ProbsRatio(dist_inputs, actions, old_log_probs, log_prob_fun)
  assert probs_ratio.shape == old_log_probs.shape, (
      f'probs_ratio.shape was {probs_ratio.shape} and'
      f'old_log_probs.shape was {old_log_probs.shape}')

  # jaxified versions of
  # returns[dones] = rewards[dones]
  # values[dones] = 0
  returns = jnp.where(dones, rewards, returns)
  values = jnp.where(dones, jnp.zeros_like(values), values)
  advantages = returns - values
  if normalize_advantages:
    advantages = advantages - jnp.mean(advantages)
    advantages /= jnp.std(advantages) + 1e-8
  assert old_log_probs.shape == advantages.shape, (
      f'old_log_probs.shape was {old_log_probs.shape} and advantages.shape was '
      f'{advantages.shape}')

  unclipped_objective = UnclippedObjective(probs_ratio, advantages)
  assert unclipped_objective.shape == advantages.shape, (
      f'old_log_probs.shape was {old_log_probs.shape} and'
      f'unclipped_objective.shape was {unclipped_objective.shape}')

  clipped_objective = ClippedObjective(probs_ratio, advantages, epsilon)
  assert clipped_objective.shape == advantages.shape, (
      f'clipped_objective.shape was {clipped_objective.shape} and'
      f'advantages.shape was {advantages.shape}')

  ppo_objective = jnp.minimum(unclipped_objective, clipped_objective)
  assert ppo_objective.shape == advantages.shape, (
      f'ppo_objective.shape was {ppo_objective.shape} and'
      f'advantages.shape was {advantages.shape}')

  return ppo_objective


def A2CObjective(dist_inputs, values, returns, dones, rewards,
                 actions, mask, log_prob_fun, normalize_advantages):
  """Definition of the Advantage Actor Critic (A2C) loss."""
  # dist_inputs of the shape float32[128,1,18]
  # values of the shape float32[128,1,1]
  # returns of the shape float32[128,1,1]
  # dones of the shape int32[128,1,1]
  # actions of the shape int32[128,1]
  # and mask of the shape float32[128,1]
  # We have to squeeze values and returns, because we
  # are planning to compute (return - values) * new_log_probs * mask
  # and all of them should be of the same dimension
  values = values.squeeze(axis=2)
  returns = returns.squeeze(axis=2)
  dones = dones.squeeze(axis=2)
  rewards = rewards.squeeze(axis=2)
  assert rewards.shape == dones.shape, (
      f'rewards.shape was {rewards.shape} and dones.shape was {dones.shape}')
  assert dones.shape == values.shape, (
      f'dones.shape was {dones.shape} and values.shape was {values.shape}')
  assert returns.shape == values.shape, (
      f'returns.shape was {returns.shape} and values.shape was {values.shape}')
  assert values.shape == mask.shape, (
      f'values.shape was {values.shape} and mask.shape was {mask.shape}')
  assert returns.shape[0] == dist_inputs.shape[0], (
      f'returns.shape[0] was {returns.shape[0]} and dist_inputs.shape[0] was '
      f'{dist_inputs.shape[0]}')

  new_log_probs = NewLogProbs(dist_inputs, actions, log_prob_fun)
  assert new_log_probs.shape == mask.shape, (
      f'new_log_probs.shape was {new_log_probs.shape} and mask.shape was '
      f'{mask.shape}')

  # jaxified versions of
  # returns[dones] = rewards[dones]
  # values[dones] = 0
  returns = jnp.where(dones, rewards, returns)
  values = jnp.where(dones, jnp.zeros_like(values), values)
  advantages = returns - values
  if normalize_advantages:
    advantages = advantages - jnp.mean(advantages)
    advantages /= jnp.std(advantages) + 1e-8
  assert new_log_probs.shape == advantages.shape, (
      f'new_log_probs.shape was {new_log_probs.shape} and advantages.shape was '
      f'{advantages.shape}')

  # One of the motivation to the squeezes and assertions is to
  # avoid [128,1] * [128,1,1] * [128] multiplications in the definition
  # of the a2c objective - we insist on the same shapes
  a2c_objective = -jnp.sum(new_log_probs * advantages * mask) / jnp.sum(mask)
  return a2c_objective



# coding=utf-8
# Copyright 2021 The Trax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests different ways to use the public tf-numpy module."""
import numpy as onp

import tensorflow as tf
import tensorflow.experimental.numpy as np1
from tensorflow.experimental import numpy as np2  # pylint: disable=reimported


np3 = tf.experimental.numpy


class PublicSymbolTest(tf.test.TestCase):

  def testSimple(self):
    a = 0.1
    b = 0.2
    for op in [np1.add, np2.add, np3.add]:
      self.assertAllClose(onp.add(a, b), op(a, b))


if __name__ == "__main__":
  tf.compat.v1.enable_eager_execution()
  tf.test.main()



# coding=utf-8
# Copyright 2021 The Trax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Policy network training tasks.

Policy tasks encapsulate the training process of a policy network into a simple,
replaceable component. To implement a policy-based Agent using policy tasks:

  1. Subclass the base Agent class.
  2. In __init__(), initialize the policy training and evaluation tasks, and
     a trax.supervised.training.Loop instance using them.
  3. In train_epoch(), call the Loop to train the network.
  4. In policy(), call network_policy() defined in this module.
"""

import numpy as np

from trax import layers as tl
from trax.fastmath import numpy as jnp
from trax.rl import distributions
from trax.supervised import training


class PolicyTrainTask(training.TrainTask):
  """Task for policy training.

  Trains the policy based on action advantages.
  """

  def __init__(
      self,
      trajectory_batch_stream,
      optimizer,
      lr_schedule,
      policy_distribution,
      advantage_estimator,
      value_fn,
      weight_fn=(lambda x: x),
      advantage_normalization=True,
      advantage_normalization_epsilon=1e-5,
      head_selector=(),
  ):
    """Initializes PolicyTrainTask.

    Args:
      trajectory_batch_stream: Generator of trax.rl.task.TimeStepBatch.
      optimizer: Optimizer for network training.
      lr_schedule: Learning rate schedule for network training.
      policy_distribution: Distribution over actions.
      advantage_estimator: Function
        (rewards, returns, values, dones) -> advantages, created by one of the
        functions from trax.rl.advantages.
      value_fn: Function TimeStepBatch -> array (batch_size, seq_len)
        calculating the baseline for advantage calculation. Can be used to
        implement actor-critic algorithms, by substituting a call to the value
        network as value_fn.
      weight_fn: Function float -> float to apply to advantages. Examples:
        - A2C: weight_fn = id
        - AWR: weight_fn = exp
        - behavioral cloning: weight_fn(_) = 1
      advantage_normalization: Whether to normalize advantages.
      advantage_normalization_epsilon: Epsilon to use then normalizing
        advantages.
      head_selector: Layer to apply to the network output to select the value
        head. Only needed in multitask training. By default, use a no-op layer,
        signified by an empty sequence of layers, ().
    """
    self.trajectory_batch_stream = trajectory_batch_stream
    self._value_fn = value_fn
    self._advantage_estimator = advantage_estimator
    self._weight_fn = weight_fn
    self._advantage_normalization = advantage_normalization
    self._advantage_normalization_epsilon = advantage_normalization_epsilon
    self.policy_distribution = policy_distribution

    labeled_data = map(self.policy_batch, trajectory_batch_stream)
    sample_batch = self.policy_batch(
        next(trajectory_batch_stream), shape_only=True
    )
    loss_layer = distributions.LogLoss(distribution=policy_distribution)
    loss_layer = tl.Serial(head_selector, loss_layer)
    super().__init__(
        labeled_data, loss_layer, optimizer,
        sample_batch=sample_batch,
        lr_schedule=lr_schedule,
        loss_name='policy_loss',
    )

  def calculate_advantages(self, trajectory_batch, shape_only=False):
    (batch_size, seq_len) = trajectory_batch.observation.shape[:2]
    assert trajectory_batch.action.shape[:2] == (batch_size, seq_len)
    assert trajectory_batch.mask.shape == (batch_size, seq_len)
    if shape_only:
      values = np.zeros((batch_size, seq_len))
    else:
      # Compute the value, i.e. baseline in advantage computation.
      values = np.array(self._value_fn(trajectory_batch))
      assert values.shape == (batch_size, seq_len)
    # Compute the advantages using the chosen advantage estimator.
    return self._advantage_estimator(
        rewards=trajectory_batch.reward,
        returns=trajectory_batch.return_,
        dones=trajectory_batch.done,
        values=values,
        discount_mask=trajectory_batch.env_info.discount_mask,
    )

  def calculate_weights(self, advantages):
    """Calculates advantage-based weights for log loss in policy training."""
    if self._advantage_normalization:
      # Normalize advantages.
      advantages -= jnp.mean(advantages)
      advantage_std = jnp.std(advantages)
      advantages /= advantage_std + self._advantage_normalization_epsilon
    weights = self._weight_fn(advantages)
    assert weights.shape == advantages.shape
    return weights

  def trim_and_mask_batch(self, trajectory_batch, advantages):
    (batch_size, seq_len) = trajectory_batch.observation.shape[:2]
    adv_seq_len = advantages.shape[1]
    # The advantage sequence should be shorter by the margin. Margin is the
    # number of timesteps added to the trajectory slice, to make the advantage
    # estimation more accurate. adv_seq_len determines the length of the target
    # sequence, and is later used to trim the inputs and targets in the training
    # batch. Example for margin 2:
    # observations.shape == (4, 5, 6)
    # rewards.shape == values.shape == (4, 5)
    # advantages.shape == (4, 3)
    assert adv_seq_len <= seq_len
    assert advantages.shape == (batch_size, adv_seq_len)
    # Trim observations, actions and mask to match the target length.
    observations = trajectory_batch.observation[:, :adv_seq_len]
    actions = trajectory_batch.action[:, :adv_seq_len]
    mask = trajectory_batch.mask[:, :adv_seq_len]
    # Apply the control mask, so we only compute policy loss for controllable
    # timesteps.
    mask *= trajectory_batch.env_info.control_mask[:, :adv_seq_len]
    return (observations, actions, mask)

  def policy_batch(self, trajectory_batch, shape_only=False):
    """Computes a policy training batch based on a trajectory batch.

    Args:
      trajectory_batch: trax.rl.task.TimeStepBatch with a batch of trajectory
        slices. Elements should have shape (batch_size, seq_len, ...).
      shape_only: Whether to return dummy zero arrays of correct shape. Useful
        for initializing models.

    Returns:
      Triple (observations, actions, weights), where weights are the
      advantage-based weights for the policy loss. Shapes:
      - observations: (batch_size, seq_len) + observation_shape
      - actions: (batch_size, seq_len) + action_shape
      - weights: (batch_size, seq_len)
    """
    advantages = self.calculate_advantages(
        trajectory_batch, shape_only=shape_only
    )
    (observations, actions, mask) = self.trim_and_mask_batch(
        trajectory_batch, advantages
    )
    weights = self.calculate_weights(advantages) * mask / jnp.sum(mask)
    return (observations, actions, weights)


class PolicyEvalTask(training.EvalTask):
  """Task for policy evaluation."""

  def __init__(self, train_task, n_eval_batches=1, head_selector=()):
    """Initializes PolicyEvalTask.

    Args:
      train_task: PolicyTrainTask used to train the policy network.
      n_eval_batches: Number of batches per evaluation.
      head_selector: Layer to apply to the network output to select the value
        head. Only needed in multitask training.
    """
    self._train_task = train_task
    self._policy_dist = train_task.policy_distribution
    labeled_data = map(self._eval_batch, train_task.trajectory_batch_stream)
    sample_batch = self._eval_batch(
        next(train_task.trajectory_batch_stream), shape_only=True
    )
    # TODO(pkozakowski): Implement more metrics.
    metrics = {
        'policy_entropy': self.entropy_metric,
    }
    metrics.update(self.advantage_metrics)
    metrics.update(self.weight_metrics)
    metrics = {
        name: tl.Serial(head_selector, metric)
        for (name, metric) in metrics.items()
    }
    (metric_names, metric_layers) = zip(*metrics.items())
    # Select the appropriate head for evaluation.
    super().__init__(
        labeled_data, metric_layers,
        sample_batch=sample_batch,
        metric_names=metric_names,
        n_eval_batches=n_eval_batches,
    )

  def _eval_batch(self, trajectory_batch, shape_only=False):
    advantages = self._train_task.calculate_advantages(
        trajectory_batch, shape_only=shape_only
    )
    (observations, actions, mask) = self._train_task.trim_and_mask_batch(
        trajectory_batch, advantages
    )
    return (observations, actions, advantages, mask)

  @property
  def entropy_metric(self):
    def Entropy(policy_inputs, actions, advantages, mask):
      del actions, advantages, mask
      return jnp.mean(self._policy_dist.entropy(policy_inputs))
    return tl.Fn('Entropy', Entropy)

  @property
  def advantage_metrics(self):
    def make_metric(aggregate_fn):  # pylint: disable=invalid-name
      def AdvantageMetric(policy_inputs, actions, advantages, mask):
        del policy_inputs, actions, mask
        return aggregate_fn(advantages)
      return tl.Fn('AdvantageMetric', AdvantageMetric)
    return {
        'advantage_' + name: make_metric(fn) for (name, fn) in [
            ('mean', jnp.mean),
            ('std', jnp.std),
        ]
    }

  @property
  def weight_metrics(self):
    def make_metric(aggregate_fn):  # pylint: disable=invalid-name
      def WeightMetric(policy_inputs, actions, advantages, mask):
        del policy_inputs, actions, mask
        weights = self._train_task.calculate_weights(advantages)
        return aggregate_fn(weights)
      return tl.Fn('WeightMetric', WeightMetric)
    return {  # pylint: disable=g-complex-comprehension
        'weight_' + name: make_metric(fn) for (name, fn) in [
            ('mean', jnp.mean),
            ('std', jnp.std),
            ('min', jnp.min),
            ('max', jnp.max),
        ]
    }



# coding=utf-8
# Copyright 2021 The Trax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Trax-to-Keras converter."""

import functools

import tensorflow.compat.v2 as tf

from trax import fastmath as math_lib
from trax import shapes as shapes_lib
from trax.fastmath import numpy as jnp
from trax.layers import base


def _replace_none_batch(x, batch_size=None):
  if batch_size is None:
    return x
  if isinstance(x, tf.Tensor) and x.shape[0] is None:
    x.set_shape([batch_size] + x.shape[1:])
    return x
  elif isinstance(x, tf.TensorShape) and x[0] is None:
    return [batch_size] + x[1:]
  return x


def tensor_shapes_to_shape_dtypes(shapes, dtype):
  return math_lib.nested_map(
      lambda s: shapes_lib.ShapeDtype(s.as_list(), dtype), shapes)


def read_values(variables):
  return math_lib.nested_map(lambda v: v.read_value(), variables)


def to_tensors(args):
  return math_lib.nested_map(tf.convert_to_tensor, args)


def to_arrays(args):
  return math_lib.nested_map(jnp.asarray, args)


class AsKeras(tf.keras.layers.Layer):
  """A Keras layer built from a Trax layer.

  This subclass of `tf.keras.layers.Layer` takes in a Trax layer as a
  constructor argument and wraps it to be a Keras layer. It uses
  `tf.Variable` to store weights and state (initialized according to the Trax
  layer), and uses the Trax layer's forward function as its forward function.

  Consider this code snippet::

      keras_layer = AsKeras(trax_layer, initializer_rng=initializer_rng,
                                   rng=rng, rng_updater=rng_updater)
      keras_layer.build(...)  # optional
      outputs = keras_layer(inputs)

  (Note that in Keras calling `Layer.build` is optional. If omitted, it will be
  called automatically by `Layer.__call__`.)

  If `trax_layer` already has weights at `build` time, the snippet is roughly
  equivalent to::

      weights = trax_layer.weights
      state = trax_layer.state
      keras_layer = tf.keras.layers.Layer()
      keras_layer._weights = tf.Variable(weights)
      keras_layer._state = tf.Variable(state)
      keras_layer._rng = tf.Variable(rng)
      outputs, new_state = trax_layer(inputs, keras_layer._weights,
                                      keras_layer._state, keras_layer._rng)
      keras_layer._state.assign(new_state)
      keras_layer._rng.assign(rng_updater(rng))

  If `trax_layer` doesn't have weights at `build` time, the snippet is roughly
  equivalent to::

      weights, state = trax_layer.init(..., rng=initializer_rng)
      keras_layer = ...
      ...

  `AsKeras` uses `tf.Variable` to store weights, not shared with the
  original Trax layer (which uses tensors to store weights), so using
  `AsKeras` may double the memory footprint. This problem can be solved
  by making sure that the Trax layer's weights/state are cleared whenever
  `tf.Variable.assign` (and `tf.Variable.assign_add` etc.) is called, because
  `tf.Variable` is copy-on-write by default.

  Mutations in those `tf.Variable`s won't affect the Trax layer's weights, but
  `AsKeras`'s forward function calls the Trax layer's forward function,
  which caches the weights in the Trax layer object, so a forward pass may
  change the weights cached in the original Trax layer.

  Note that this class is not thread-safe. If the same `AsKeras` object
  is used in multiple threads, the `tf.Variable` updates may happen in a
  non-deterministic order.
  """

  def __init__(self, trax_layer, batch_size=None, initializer_rng=None,
               rng=None, rng_updater=None, dtype=None):
    """Creates a Keras layer wrapping around a Trax layer.

    Args:
      trax_layer: an object of class `trax.layers.Layer`, the trax layer to
        wrap.
      batch_size: (optional) an integer, the batch size that this Keras layer
        will be used on. Keras sometimes needs to generate a TF graph for a
        layer (e.g. for acceleration or checkpointing). The inputs used to trace
        the graph will have `None` as the length of their batch dimensions, so
        as to generate a graph that can handle any batch size. Some Trax layers
        can't handle tensors whose shapes contain `None`. If `batch_size` is set
        to an integer, the graph will be traced with `batch_size` as the batch
        size instead of `None`. Note that in this case the graph (and the Keras
        layer) can only be used on a specific batch size. If you want to use a
        different batch size, you need to create another `AsKeras` object
        with a different `batch_size`.
      initializer_rng: (optional) an RNG key used to create the weights and
        state if `trax_layer` doesn't have them. If `None`,
        `trax.fastmath.random.get_prng(0)` will be used.
      rng: (optional) an RNG key for the forward function (aka the "forward
        key"). If `None`, `trax.fastmath.random.get_prng(0)` will be used.
      rng_updater: (optional) a function of type rng_key -> rng_key, used to
        update the forward key after each forward pass. If `None`, the function
        `lambda x: trax.fastmath.random.split(x, 1)[0]` will be used, which
        advances the RNG key.
      dtype: (optional) the dtype of the inputs. See the `dtype` argument of
        `tf.keras.layers.Layer.__init__` for details.
    """
    super().__init__(dtype=dtype)
    with math_lib.use_backend(math_lib.Backend.TFNP):
      if initializer_rng is None:
        initializer_rng = math_lib.random.get_prng(0)
      if rng is None:
        rng = math_lib.random.get_prng(0)
      if rng_updater is None:
        rng_updater = lambda x: math_lib.random.split(x, 1)[0]
      self._trax_layer = trax_layer
      self._batch_size = batch_size
      self._initializer_rng = initializer_rng
      self._forward_rng_init = rng
      self._rng_updater = rng_updater

  def build(self, input_shape):
    with math_lib.use_backend(math_lib.Backend.TFNP):
      # Using `is` instead of `==` following Trax's practice
      if self._trax_layer.weights is base.EMPTY_WEIGHTS:
        sanitized_input_shape = math_lib.nested_map(
            functools.partial(_replace_none_batch, batch_size=self._batch_size),
            input_shape)
        weights, state = self._trax_layer.init(
            tensor_shapes_to_shape_dtypes(sanitized_input_shape, self.dtype),
            rng=self._initializer_rng)
      else:
        weights = self._trax_layer.weights
        state = self._trax_layer.state
      # Note: `weights` may contain `EMPTY_WEIGHTS`
      self._weights = math_lib.nested_map(
          functools.partial(tf.Variable, trainable=True), weights)
      self._state = math_lib.nested_map(
          functools.partial(tf.Variable, trainable=False), state)
      self._rng = tf.Variable(self._forward_rng_init, trainable=False)
    super().build(input_shape)

  def call(self, inputs):
    with math_lib.use_backend(math_lib.Backend.TFNP):
      inputs = math_lib.nested_map(
          functools.partial(_replace_none_batch, batch_size=self._batch_size),
          inputs)
      weights, state, rng = read_values([self._weights, self._state, self._rng])
      inputs, weights, state, rng = to_arrays([inputs, weights, state, rng])
      outputs, new_state = self._trax_layer.pure_fn(inputs, weights=weights,
                                                    state=state, rng=rng)
      tf.nest.map_structure(lambda v, t: v.assign(t), self._state, new_state)
      self._rng.assign(self._rng_updater(rng))
      outputs = to_tensors(outputs)
      return outputs



# coding=utf-8
# Copyright 2021 The Trax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Trax fast math: JAX backend."""

import functools
import jax
from jax import lax
from jax import random as jax_random
import jax.numpy as jnp
import jax.scipy.special as jax_special
import numpy as np
import tensorflow as tf
import tensorflow_datasets as tfds

from trax.fastmath import numpy as tnp
from trax.shapes import signature


def jax_conv(inp, fltr, window_strides, padding, dimension_numbers,
             filter_dilation=None):
  """A wrapper around `lax.conv_general_dilated`.

  It requires `dimension_numbers` and disallows `inp_dilation`.

  Args:
    inp: an (N+2)-D array. The input of the convolution.
    fltr: an (N+2)-D array. The filter (i.e. kernel) of the convolution.
    window_strides: the strides for moving the convolution window.
    padding: a string, either 'VALID' or 'SAME'. The padding algorithm.
    dimension_numbers: a tuple of three strings encoding the data format of
      input, filter and output. 'I' means input; 'O' means output; 'C' means
      channel; other characters such as 'W', 'H' and 'D' means spatial
      dimensions.
    filter_dilation: the dilation rates for the filter. Dilating the filter
      means adding "holes" to the filter.

  Returns:
    An (N+2)-D array. The convolution result.
  """
  return lax.conv_general_dilated(inp, fltr, window_strides, padding,
                                  lhs_dilation=None,
                                  rhs_dilation=filter_dilation,
                                  dimension_numbers=dimension_numbers)


def _pooling_general(inputs, reducer, init_val, rescaler=None,
                     pool_size=(2, 2), strides=None, padding='VALID'):
  """Helper: general pooling computation used in pooling layers later."""
  spatial_strides = strides or (1,) * len(pool_size)
  rescale = rescaler(pool_size, spatial_strides, padding) if rescaler else None
  dims = (1,) + pool_size + (1,)  # NHWC
  strides = (1,) + spatial_strides + (1,)
  out = lax.reduce_window(inputs, init_val, reducer, dims, strides, padding)
  return rescale(out, inputs) if rescale else out  # pylint: disable=not-callable


def jax_max_pool(x, pool_size, strides, padding):
  return _pooling_general(x, lax.max, -jnp.inf, pool_size=pool_size,
                          strides=strides, padding=padding)


def jax_sum_pool(x, pool_size, strides, padding):
  return _pooling_general(x, lax.add, 0., pool_size=pool_size,
                          strides=strides, padding=padding)


def _normalize_by_window_size(dims, spatial_strides, padding):  # pylint: disable=invalid-name
  def rescale(outputs, inputs):
    one = jnp.ones(inputs.shape[1:-1], dtype=inputs.dtype)
    window_sizes = lax.reduce_window(
        one, 0., lax.add, dims, spatial_strides, padding)
    return outputs / window_sizes[..., jnp.newaxis]
  return rescale


def jax_avg_pool(x, pool_size, strides, padding):
  return _pooling_general(x, lax.add, 0., _normalize_by_window_size,
                          pool_size, strides=strides, padding=padding)


def jax_abstract_eval(f):
  """Returns a function that evaluates `f` given input shapes and dtypes.

  It transforms function `f` to a function that performs the same computation as
  `f` but only on shapes and dtypes (a.k.a. shape inference).

  Args:
    f: the function to be transformed.

  Returns:
    A function whose input arguments can be either the same as `f`'s or only
    their shapes/dtypes represented by `ShapeDtype`, and whose return values are
    `ShapeDtype`s with the same nested structure as `f`'s return values.
  """
  def shape_fun(*args, **kwargs):
    jax_shapes = jax.eval_shape(f, *args, **kwargs)
    return tnp.nested_map(signature, jax_shapes)
  return shape_fun


# The default value of dtype is different from jax_random.randint
def jax_randint(key, shape, minval, maxval, dtype=np.int32):
  """Sample uniform random values in [minval, maxval) with given shape/dtype.

  Args:
    key: a PRNGKey used as the random key.
    shape: a tuple of nonnegative integers representing the shape.
    minval: int or array of ints broadcast-compatible with ``shape``, a minimum
      (inclusive) value for the range.
    maxval: int or array of ints broadcast-compatible with  ``shape``, a maximum
      (exclusive) value for the range.
    dtype: optional, an int dtype for the returned values (default int32).

  Returns:
    A random array with the specified shape and dtype.
  """
  return jax_random.randint(key, shape, minval=minval, maxval=maxval,
                            dtype=dtype)


def _to_numpy(x):
  """Converts non-NumPy tensors to NumPy arrays."""
  return x if isinstance(x, np.ndarray) else x.numpy()


def _dataset_as_numpy(ds, batch_size=None):
  """Speed up tfds.as_numpy by batching and then iterating over the batches."""
  batch_size = batch_size or 1
  try:  # Check that dense_to_ragged_batch exists.
    if batch_size < 2:  # Fall back to default if no batching requested.
      raise AttributeError
    ds_batch = ds.apply(tf.data.experimental.dense_to_ragged_batch(batch_size))
    for example in tfds.as_numpy(ds_batch):
      flat_example = tnp.tree_flatten(example)
      np_flat_example = [_to_numpy(x) for x in flat_example]
      for single_example_flat in zip(*np_flat_example):
        single_example, _ = tnp.tree_unflatten(single_example_flat, example)
        yield single_example
  except AttributeError:
    # In TF 1.X there is not dense_to_ragged_batch: fallback.
    for example in tfds.as_numpy(ds):
      yield example


def _custom_grad(f_vjp, f_original):
  f_ = jax.custom_transforms(f_original)
  jax.defvjp_all(f_, f_vjp)
  return f_


def _custom_vjp(f, f_fwd, f_bwd, nondiff_argnums=()):
  @functools.partial(jax.custom_vjp, nondiff_argnums=nondiff_argnums)
  def _f(*args, **kwargs):
    return f(*args, **kwargs)
  _f.defvjp(f_fwd, f_bwd)
  return _f


JAX_BACKEND = {
    'name': 'jax',
    'np': jnp,
    'abstract_eval': jax_abstract_eval,
    'avg_pool': jax_avg_pool,
    'cond': lax.cond,
    'conv': jax_conv,
    'custom_vjp': _custom_vjp,
    'custom_grad': _custom_grad,
    'dataset_as_numpy': _dataset_as_numpy,
    'dynamic_slice': jax.lax.dynamic_slice,
    'dynamic_slice_in_dim': jax.lax.dynamic_slice_in_dim,
    'dynamic_update_slice': jax.lax.dynamic_update_slice,
    'dynamic_update_slice_in_dim': jax.lax.dynamic_update_slice_in_dim,
    'erf': jax_special.erf,
    'expit': jax_special.expit,
    'fori_loop': lax.fori_loop,
    'global_device_count': jax.device_count,
    'grad': jax.grad,
    'value_and_grad': jax.value_and_grad,
    'index_add': lambda x, idx, y: jnp.asarray(x).at[idx].add(y),
    'index_max': lambda x, idx, y: jnp.asarray(x).at[idx].max(y),
    'index_min': lambda x, idx, y: jnp.asarray(x).at[idx].min(y),
    'index_update': lambda x, idx, y: jnp.asarray(x).at[idx].set(y),
    'jit': jax.jit,
    'local_device_count': jax.local_device_count,
    'logsumexp': jax_special.logsumexp,
    'lt': lax.lt,
    'map': lax.map,
    'max_pool': jax_max_pool,
    'pmap': jax.pmap,
    'psum': lax.psum,
    'random_bernoulli': jax_random.bernoulli,
    'random_get_prng': jax.jit(jax_random.PRNGKey),
    'random_normal': jax_random.normal,
    'random_randint': jax_randint,
    'random_split': jax_random.split,
    'random_fold_in': jax_random.fold_in,
    'random_uniform': jax_random.uniform,
    'remat': jax.remat,
    'scan': lax.scan,
    'sort_key_val': jax.lax.sort_key_val,
    'stop_gradient': lax.stop_gradient,
    'sum_pool': jax_sum_pool,
    'top_k': lax.top_k,
    'vjp': jax.vjp,
    'vmap': jax.vmap,
}



# coding=utf-8
# Copyright 2021 The Trax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Transformer variant -- no encoder-decoder attention."""

from trax import fastmath
from trax import layers as tl
from trax.fastmath import numpy as jnp
from trax.models.research import configurable_transformer as ct


def Transformer2(input_vocab_size,
                 output_vocab_size=None,
                 d_model=512,
                 d_ff=2048,
                 n_encoder_layers=6,
                 n_decoder_layers=6,
                 n_heads=8,
                 dropout=0.1,
                 dropout_shared_axes=None,
                 max_len=2048,
                 mode='train',
                 ff_activation=tl.Relu,
                 ff_dropout=0.1,
                 ff_chunk_size=0,
                 ff_use_sru=0,
                 ff_sparsity=0,
                 ff_sparsity_type='1inN',
                 attention_chunk_size=0,
                 encoder_attention_type=tl.Attention,
                 n_encoder_attention_layers=1,
                 decoder_attention_type=tl.CausalAttention,
                 n_decoder_attention_layers=2,
                 pos_type=None,
                 pos_axial_shape=None,
                 pos_d_axial_embs=None):
  """Returns a Transformer model.

  This model expects an input pair: target, source.

  Args:
    input_vocab_size: int: vocab size of the source.
    output_vocab_size: int (optional): vocab size of the target. If None, the
      source and target are assumed to have the same vocab.
    d_model: int:  depth of embedding
    d_ff: int: depth of feed-forward layer
    n_encoder_layers: int: number of encoder layers
    n_decoder_layers: int: number of decoder layers
    n_heads: int: number of attention heads
    dropout: float: dropout rate (how much to drop out)
    dropout_shared_axes: axes on which to share dropout mask
    max_len: int: maximum symbol length for positional encoding
    mode: str: 'train' or 'eval'
    ff_activation: the non-linearity in feed-forward layer
    ff_dropout: Stochastic rate (probability) for dropping an activation value
      when applying dropout after the FF dense layer.
    ff_chunk_size: int; if > 0, chunk feed-forward into this-sized chunks
    ff_use_sru: int; if > 0, we use this many SRU layers instead of feed-forward
    ff_sparsity: int, if > 0 use sparse feed-forward block with this sparsity
    ff_sparsity_type: string, if ff_sparsity >0,
      use SparseFF if ff_sparsity_type=`'1inN'` and
      use BlockSparseFF if ff_sparsity_type=`'Block'`
    attention_chunk_size: int, if > 0 run attention chunked at this size
    encoder_attention_type: The attention layer to use for the encoder part.
    n_encoder_attention_layers: int, within each encoder block, how many
      attention layers to have.
    decoder_attention_type: The attention layer to use for the
      encoder-decoder attention.
    n_decoder_attention_layers: int, within each decoder block, how many
      attention layers to have.
    pos_type: string, the type of positional embeddings to use.
    pos_axial_shape: tuple of ints: input shape to use for the axial position
      encoding. If unset, axial position encoding is disabled.
    pos_d_axial_embs: tuple of ints: depth of position embedding for each axis.
      Tuple length must match pos_axial_shape, and values must sum to d_model.

  Returns:
    A Transformer model as a layer that maps from a target, source pair to
    activations over a vocab set.
  """
  in_encoder, out_encoder, output_vocab_size = (
      ct.EmbeddingAndPositionalEncodings(
          input_vocab_size,
          d_model,
          mode,
          dropout,
          dropout_shared_axes,
          max_len,
          output_vocab_size=output_vocab_size,
          pos_type=pos_type,
          pos_axial_shape=pos_axial_shape,
          pos_d_axial_embs=pos_d_axial_embs)
  )

  # pylint: disable=g-complex-comprehension
  encoder_blocks = [
      ct.EncoderBlock(d_model, d_ff, n_heads, dropout, dropout_shared_axes,
                      mode, ff_activation, ff_dropout, ff_chunk_size,
                      ff_use_sru, ff_sparsity, ff_sparsity_type,
                      attention_chunk_size, encoder_attention_type,
                      n_encoder_attention_layers)
      for i in range(n_encoder_layers)]
  # pylint: enable=g-complex-comprehension

  encoder = tl.Serial(
      in_encoder,
      encoder_blocks,
      tl.LayerNorm()
  )
  if mode == 'predict':
    encoder = tl.Cache(encoder)

  # pylint: disable=g-complex-comprehension
  decoder_blocks = [
      ct.DecoderBlock(d_model, d_ff, n_heads, dropout, dropout_shared_axes,
                      mode, ff_activation, ff_dropout, ff_chunk_size,
                      ff_use_sru, ff_sparsity, ff_sparsity_type,
                      attention_chunk_size, decoder_attention_type,
                      n_decoder_attention_layers)
      for i in range(n_decoder_layers)]
  # pylint: enable=g-complex-comprehension

  # Assemble and return the model.
  return tl.Serial(
      # Input: encoder_side_tokens, decoder_side_tokens
      # Copy decoder tokens for use in loss.
      tl.Select([0, 0, 1, 1]),          # tok_e tok_e tok_d tok_d

      # Encode.
      tl.Branch([], tl.PaddingMask()),  # tok_e mask_e tok_e tok_d tok_d
      encoder,                          # vec_e mask_e tok_e tok_d tok_d

      # Simple encoder mask, doesn't contain extra dims.
      tl.Select([2, 0, 2], n_in=3),     #  tok_e vec_e tok_e tok_d tok_d
      tl.Fn('EncoderMask',              # mask_e vec_e tok_e tok_d tok_d
            lambda x: x != 0, n_out=1),

      # Decode.
      tl.Select([3, 1, 0, 2]),          #  tok_d vec_e mask_e tok_e tok_d
      tl.ShiftRight(mode=mode),         # stok_d vec_e mask_e tok_e tok_d
      out_encoder,                      # svec_d vec_e mask_e tok_e tok_d

      # Concat encoder and decoder.
      tl.Select([1, 0]),                # vec_e svec_d mask_e tok_e tok_d
      ConcatWithPadding(mode=mode),     # vec_ed tok_e tok_d

      # Decoder blocks with causal attention
      decoder_blocks,                   # vec_ed tok_e tok_d
      tl.LayerNorm(),                   # vec_ed tok_e tok_d

      # Separate out the encoder part from the concatenated vector.
      tl.Select([0, 1, 2, 2]),                     # vec_ed tok_e tok_d tok_d
      StripFromConcatenateWithPadding(mode=mode),  # vec_d tok_d

      # Map to output vocab.
      tl.Dense(output_vocab_size),      # vec_d tok_d
  )


# Arg shapes: (B, L1, H), (B, L2, H), (B, L1).
def _ConcatWithPadding(vec_e, vec_d, mask_e):
  """Concatenate with padding: see the ConcatWithPadding layer for details."""
  # pylint: disable=invalid-name
  B, L1, H = vec_e.shape
  L2 = vec_d.shape[1]
  # pylint: enable=invalid-name

  if vec_d.shape != (B, L2, H):
    raise ValueError(f'Shape of decoder vector, {vec_d.shape}, does not'
                     f' equal {(B, L2, H)}.')
  if mask_e.shape != (B, L1):
    raise ValueError(f'Shape of encoder mask, {mask_e.shape}, does not'
                     f' equal {(B, L1)}.')

  def _UpdateRow(x):
    # row_e - (L1, H), row_d - (L2, H), row_mask_e - (L1,)
    row_e, row_d, row_mask_e = x
    # final_row - (L1+L2, H)
    final_row = jnp.concatenate([row_e, jnp.zeros_like(row_d)], axis=0)
    # Find the last real token/vector of the encoder.
    e_idx = jnp.sum(row_mask_e, dtype=jnp.int32)
    # Starting after that index, update with the decoder row.
    zero = jnp.array(0, dtype=e_idx.dtype)  # avoid int32/int64 mismatch
    return fastmath.dynamic_update_slice(final_row, row_d, (e_idx, zero))

  return fastmath.map(_UpdateRow, [vec_e, vec_d, mask_e])


def _StripFromConcatenateWithPadding(vec_ed, tok_e, tok_d):
  """Strip concatenate with padding: see the layer below for details."""
  # pylint: disable=invalid-name
  B, L, H = vec_ed.shape
  L1 = tok_e.shape[1]
  L2 = tok_d.shape[1]
  # pylint: enable=invalid-name
  if L != L1 + L2:
    raise ValueError(f'Length from encoder-decoder vectors ({L}) does not'
                     f' equal sum of lengths from encoder ({L1}) and decoder'
                     f' ({L2}).')
  if tok_e.shape != (B, L1):
    raise ValueError(f'Shape of encoder tokens, {tok_e.shape}, does not'
                     f' equal {(B, L1)}.')
  if tok_d.shape != (B, L2):
    raise ValueError(f'Shape of decoder tokens, {tok_d.shape}, does not'
                     f' equal {(B, L2)}.')

  def _UpdateRow(x):
    # (L, H), (L1, H) & (L2, H)
    row_ed, row_e, _ = x
    mask_e = row_e != 0
    len_e = jnp.sum(mask_e, dtype=jnp.int32)
    # In `row_ed` start where encoder tokens/vecs end, i.e. are index `len_e`
    # and pick up (L2, H) tensor slice from there.
    zero = jnp.array(0, dtype=len_e.dtype)  # avoid int32/int64 mismatch
    return fastmath.dynamic_slice(row_ed, (len_e, zero), (L2, H))

  return fastmath.map(_UpdateRow, [vec_ed, tok_e, tok_d])


class StripFromConcatenateWithPadding(tl.Layer):
  """Strips out the leading encoder tokens from the concatenated array."""

  def __init__(self, mode='train'):
    super().__init__(n_in=3, n_out=1)
    self._mode = mode

  def init_weights_and_state(self, input_signature):
    """Sets layer-specific internal state."""
    del input_signature
    self.state = jnp.array(0, dtype=jnp.int32)

  def forward(self, inputs):
    vec_ed, tok_e, tok_d = inputs

    # In training/eval mode or at the first step predict mode i.e. when
    # state.shape is (), i.e. at first step, we do the actual compuration
    if self._mode != 'predict' or not self.state.shape:
      # Now state.shape will not evaluate to false.
      self.state = self.state.reshape((1,))
      return _StripFromConcatenateWithPadding(vec_ed, tok_e, tok_d)

    # In predict mode and on subsequent steps (i.e. after the first step) vec_ed
    # is actually vec_d, since no concatenation happened at all.
    return vec_ed


class ConcatWithPadding(tl.ReversibleLayer):
  """Concatenates two length padded (B, L, H) arrays (of different lenghts)."""

  def __init__(self, mode='train'):
    super().__init__(n_in=5, n_out=3)
    self._mode = mode

  def init_weights_and_state(self, input_signature):
    """Sets layer-specific internal state."""
    del input_signature
    self.state = jnp.array(0, dtype=jnp.int32)

  def forward(self, inputs):
    vec_e, vec_d, mask_e, tok_e, tok_d = inputs

    # In training/eval mode or at the first step predict mode i.e. when
    # state.shape is (), i.e. at first step, we return the concatenated output.
    if self._mode != 'predict' or not self.state.shape:
      # Now state.shape will not evaluate to false.
      self.state = self.state.reshape((1,))
      return _ConcatWithPadding(vec_e, vec_d, mask_e), tok_e, tok_d

    # In predict mode and on subsequent steps (i.e. after the first step) we
    # don't concatenate anymore, but just return the decoder vector.
    return vec_d, tok_e, tok_d

  def reverse(self, output, weights=(), state=(), new_state=(), rng=None):
    del state, new_state, rng, weights
    assert self._mode != 'predict', 'cannot reverse in predict mode'
    vecs_ed, toks_e, toks_d = output
    vecs_d = _StripFromConcatenateWithPadding(vecs_ed, toks_e, toks_d)
    mask_e = (toks_e != 0)
    mask_e_float = mask_e.astype(jnp.float32)
    vecs_e = vecs_ed[:, :toks_e.shape[1], :] * mask_e_float[:, :, None]
    return vecs_e, vecs_d, mask_e, toks_e, toks_d


class ConcatWithPadding2(tl.ReversibleLayer):
  """Concatenate with padding operating on pairs to combine with rev-nets."""

  def __init__(self, mode='train'):
    super().__init__(n_in=6, n_out=4)
    self._mode = mode

  def init_weights_and_state(self, input_signature):
    """Sets layer-specific internal state."""
    del input_signature
    self.state = jnp.array(0, dtype=jnp.int32)

  def forward(self, inputs):
    vecs_e1, vecs_e2, vecs_d, mask_e, toks_e, toks_d = inputs

    # In training/eval mode or at the first step predict mode i.e. when
    # state.shape is (), i.e. at first step, we return the concatenated output.
    if self._mode != 'predict' or not self.state.shape:
      # Now state.shape will not evaluate to false.
      self.state = self.state.reshape((1,))
      # Calculate mask and concat_with_padding on the pairs.
      vecs_ed1 = _ConcatWithPadding(vecs_e1, vecs_d, mask_e)
      vecs_ed2 = _ConcatWithPadding(vecs_e2, vecs_d, mask_e)
      return vecs_ed1, vecs_ed2, toks_e, toks_d

    # In predict mode and on subsequent steps (i.e. after the first step) we
    # don't concatenate anymore, but just return the decoder vector.
    return vecs_d, vecs_d, toks_e, toks_d

  def reverse(self, output, weights=(), state=(), new_state=(), rng=None):
    del state, new_state, rng, weights
    assert self._mode != 'predict', 'cannot reverse in predict mode'
    vecs_ed1, vecs_ed2, toks_e, toks_d = output
    vecs_d = _StripFromConcatenateWithPadding(vecs_ed1, toks_e, toks_d)
    mask_e = (toks_e != 0)
    mask_e_float = mask_e.astype(jnp.float32)
    vecs_e1 = vecs_ed1[:, :toks_e.shape[1], :] * mask_e_float[:, :, None]
    vecs_e2 = vecs_ed2[:, :toks_e.shape[1], :] * mask_e_float[:, :, None]
    return vecs_e1, vecs_e2, vecs_d, mask_e, toks_e, toks_d



# coding=utf-8
# Copyright 2021 The Trax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Tests for trax.models.atari_cnn."""

import functools
import operator as op
import numpy as np
from tensorflow import test
from trax.models import atari_cnn
from trax.shapes import ShapeDtype


class AtariCnnTest(test.TestCase):

  def test_computes(self):
    hidden_size = (4, 4)
    output_size = 6
    model = atari_cnn.AtariCnn(
        hidden_sizes=hidden_size, output_size=output_size)
    B, T, OBS = 2, 2, (28, 28, 3)  # pylint: disable=invalid-name
    input_signature = ShapeDtype((1, 1) + OBS)
    _, _ = model.init(input_signature)
    x = np.arange(B * (T + 1) * functools.reduce(op.mul, OBS)).reshape(
        B, T + 1, *OBS)
    y = model(x)
    self.assertEqual((B, T + 1, output_size), y.shape)


class FrameStackMLPTest(test.TestCase):

  def test_computes(self):
    hidden_size = (4, 4)
    output_size = 6
    model = atari_cnn.FrameStackMLP(
        hidden_sizes=hidden_size, output_size=output_size)
    B, T, OBS = 2, 2, 3  # pylint: disable=invalid-name
    input_signature = ShapeDtype((1, 1, OBS))
    _, _ = model.init(input_signature)
    x = np.arange(B * (T + 1) * OBS).reshape(B, T + 1, OBS)
    y = model(x)
    self.assertEqual((B, T + 1, output_size), y.shape)


if __name__ == '__main__':
  test.main()



# coding=utf-8
# Copyright 2021 The Trax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""ResNet."""

from trax import layers as tl


def ConvBlock(kernel_size, filters, strides, norm, non_linearity,
              mode='train'):
  """ResNet convolutional striding block."""
  ks = kernel_size
  filters1, filters2, filters3 = filters
  main = [
      tl.Conv(filters1, (1, 1), strides),
      norm(mode=mode),
      non_linearity(),
      tl.Conv(filters2, (ks, ks), padding='SAME'),
      norm(mode=mode),
      non_linearity(),
      tl.Conv(filters3, (1, 1)),
      norm(mode=mode),
  ]
  shortcut = [
      tl.Conv(filters3, (1, 1), strides),
      norm(mode=mode),
  ]
  return [
      tl.Residual(main, shortcut=shortcut),
      non_linearity()
  ]


def IdentityBlock(kernel_size, filters, norm, non_linearity,
                  mode='train'):
  """ResNet identical size block."""
  ks = kernel_size
  filters1, filters2, filters3 = filters
  main = [
      tl.Conv(filters1, (1, 1)),
      norm(mode=mode),
      non_linearity(),
      tl.Conv(filters2, (ks, ks), padding='SAME'),
      norm(mode=mode),
      non_linearity(),
      tl.Conv(filters3, (1, 1)),
      norm(mode=mode),
  ]
  return [
      tl.Residual(main),
      non_linearity(),
  ]


def Resnet50(d_hidden=64, n_output_classes=1001, mode='train',
             norm=tl.BatchNorm,
             non_linearity=tl.Relu):
  """ResNet.

  Args:
    d_hidden: Dimensionality of the first hidden layer (multiplied later).
    n_output_classes: Number of distinct output classes.
    mode: Whether we are training or evaluating or doing inference.
    norm: `Layer` used for normalization, Ex: BatchNorm or
      FilterResponseNorm.
    non_linearity: `Layer` used as a non-linearity, Ex: If norm is
      BatchNorm then this is a Relu, otherwise for FilterResponseNorm this
      should be ThresholdedLinearUnit.

  Returns:
    The list of layers comprising a ResNet model with the given parameters.
  """

  # A ConvBlock configured with the given norm, non-linearity and mode.
  def Resnet50ConvBlock(filter_multiplier=1, strides=(2, 2)):
    filters = (
        [filter_multiplier * dim for dim in [d_hidden, d_hidden, 4 * d_hidden]])
    return ConvBlock(3, filters, strides, norm, non_linearity, mode)

  # Same as above for IdentityBlock.
  def Resnet50IdentityBlock(filter_multiplier=1):
    filters = (
        [filter_multiplier * dim for dim in [d_hidden, d_hidden, 4 * d_hidden]])
    return IdentityBlock(3, filters, norm, non_linearity, mode)

  return tl.Serial(
      tl.ToFloat(),
      tl.Conv(d_hidden, (7, 7), (2, 2), 'SAME'),
      norm(mode=mode),
      non_linearity(),
      tl.MaxPool(pool_size=(3, 3), strides=(2, 2)),
      Resnet50ConvBlock(strides=(1, 1)),
      [Resnet50IdentityBlock() for _ in range(2)],
      Resnet50ConvBlock(2),
      [Resnet50IdentityBlock(2) for _ in range(3)],
      Resnet50ConvBlock(4),
      [Resnet50IdentityBlock(4) for _ in range(5)],
      Resnet50ConvBlock(8),
      [Resnet50IdentityBlock(8) for _ in range(2)],
      tl.AvgPool(pool_size=(7, 7)),
      tl.Flatten(),
      tl.Dense(n_output_classes),
  )


def WideResnetBlock(channels, strides=(1, 1), bn_momentum=0.9, mode='train'):
  """WideResnet convolutional block."""
  return [
      tl.BatchNorm(momentum=bn_momentum, mode=mode),
      tl.Relu(),
      tl.Conv(channels, (3, 3), strides, padding='SAME'),
      tl.BatchNorm(momentum=bn_momentum, mode=mode),
      tl.Relu(),
      tl.Conv(channels, (3, 3), padding='SAME'),
  ]


def WideResnetGroup(n, channels, strides=(1, 1), bn_momentum=0.9, mode='train'):
  shortcut = [
      tl.Conv(channels, (3, 3), strides, padding='SAME'),
  ]
  return [
      tl.Residual(WideResnetBlock(channels, strides, bn_momentum=bn_momentum,
                                  mode=mode),
                  shortcut=shortcut),
      tl.Residual([WideResnetBlock(channels, (1, 1), bn_momentum=bn_momentum,
                                   mode=mode)
                   for _ in range(n - 1)]),
  ]


def WideResnet(n_blocks=3, widen_factor=1, n_output_classes=10, bn_momentum=0.9,
               mode='train'):
  """WideResnet from https://arxiv.org/pdf/1605.07146.pdf.

  Args:
    n_blocks: int, number of blocks in a group. total layers = 6n + 4.
    widen_factor: int, widening factor of each group. k=1 is vanilla resnet.
    n_output_classes: int, number of distinct output classes.
    bn_momentum: float, momentum in BatchNorm.
    mode: Whether we are training or evaluating or doing inference.

  Returns:
    The list of layers comprising a WideResnet model with the given parameters.
  """
  return tl.Serial(
      tl.ToFloat(),
      tl.Conv(16, (3, 3), padding='SAME'),
      WideResnetGroup(n_blocks, 16 * widen_factor, bn_momentum=bn_momentum,
                      mode=mode),
      WideResnetGroup(n_blocks, 32 * widen_factor, (2, 2),
                      bn_momentum=bn_momentum, mode=mode),
      WideResnetGroup(n_blocks, 64 * widen_factor, (2, 2),
                      bn_momentum=bn_momentum, mode=mode),
      tl.BatchNorm(momentum=bn_momentum, mode=mode),
      tl.Relu(),
      tl.AvgPool(pool_size=(8, 8)),
      tl.Flatten(),
      tl.Dense(n_output_classes),
  )



# coding=utf-8
# Copyright 2021 The Trax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""Trainer for RL environments.

For now we only support PPO as RL algorithm.

Sample invocation:

.. code-block:: bash

    TRAIN_BATCH_SIZE=32
    python trax/rl_trainer.py \
      --config_file=trax/rl/configs/ppo_acrobot.gin \
      --train_batch_size=${TRAIN_BATCH_SIZE} \
      --output_dir=${HOME}/ppo_acrobot \
      --alsologtostderr
"""

import faulthandler

from absl import app
from absl import flags
from absl import logging
import gin
import jax
from jax.config import config
from trax import fastmath
from trax import rl  # pylint: disable=unused-import
from trax import trainer_flags  # pylint: disable=unused-import
from trax.rl import task as rl_task
from trax.rl import training as light_trainers
from trax.tf_numpy import numpy as tf_np


FLAGS = flags.FLAGS


# Not just 'train' to avoid a conflict with trax.train in GIN files.
@gin.configurable(denylist=['output_dir'], module='trax')
def train_rl(
    output_dir,
    n_epochs=10000,
    light_rl=True,
    light_rl_trainer=light_trainers.PolicyGradient):
  """Train the RL agent.

  Args:
    output_dir: Output directory.
    n_epochs: Number epochs to run the training for.
    light_rl: deprecated, always True, left out for old gin configs.
    light_rl_trainer: which light RL trainer to use (experimental).
  """
  del light_rl
  tf_np.set_allow_float64(FLAGS.tf_allow_float64)
  task = rl_task.RLTask()
  env_name = task.env_name


  if FLAGS.jax_debug_nans:
    config.update('jax_debug_nans', True)

  if FLAGS.use_tpu:
    config.update('jax_platform_name', 'tpu')
  else:
    config.update('jax_platform_name', '')


  trainer = light_rl_trainer(task=task, output_dir=output_dir)
  def light_training_loop():
    """Run the trainer for n_epochs and call close on it."""
    try:
      logging.info('Starting RL training for %d epochs.', n_epochs)
      trainer.run(n_epochs, n_epochs_is_total_epochs=True)
      logging.info('Completed RL training for %d epochs.', n_epochs)
      trainer.close()
      logging.info('Trainer is now closed.')
    except Exception as e:
      raise e
    finally:
      logging.info('Encountered an exception, still calling trainer.close()')
      trainer.close()
      logging.info('Trainer is now closed.')

  if FLAGS.jax_debug_nans or FLAGS.disable_jit:
    fastmath.disable_jit()
    with jax.disable_jit():
      light_training_loop()
  else:
    light_training_loop()


def main(argv):
  del argv
  logging.info('Starting RL training.')

  gin_configs = FLAGS.config if FLAGS.config is not None else []
  gin.enter_interactive_mode()
  gin.parse_config_files_and_bindings(FLAGS.config_file, gin_configs)
  gin.exit_interactive_mode()

  logging.info('Gin config:')
  logging.info(gin_configs)

  train_rl(output_dir=FLAGS.output_dir)

  # TODO(afrozm): This is for debugging.
  logging.info('Dumping stack traces of all stacks.')
  faulthandler.dump_traceback(all_threads=True)

  logging.info('Training is done, should exit.')


if __name__ == '__main__':
  app.run(main)



# coding=utf-8
# Copyright 2021 The Trax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Trax convolution layers."""

import functools
import itertools
import operator

from trax import fastmath
from trax.fastmath import numpy as jnp
from trax.layers import base
from trax.layers import initializers as init


class Conv(base.Layer):
  """Layer constructor function for a general convolution layer."""

  def __init__(self, filters, kernel_size, strides=None, padding='VALID',
               dimension_numbers=('NHWC', 'HWIO', 'NHWC'),
               kernel_initializer=None,
               bias_initializer=init.RandomNormalInitializer(1e-6),
               use_bias=True):
    super().__init__()
    self._filters = filters
    self._kernel_size = kernel_size
    self._padding = padding
    self._dimension_numbers = dimension_numbers
    self._lhs_spec, self._rhs_spec, self._out_spec = dimension_numbers
    self._one = (1,) * len(kernel_size)
    self._strides = strides or self._one
    self._bias_initializer = bias_initializer
    self._use_bias = use_bias
    rhs_spec = self._rhs_spec
    self._kernel_initializer = kernel_initializer
    if kernel_initializer is None:
      self._kernel_initializer = init.GlorotNormalInitializer(
          rhs_spec.index('O'), rhs_spec.index('I'))

  def _check_nhwc(self):
    msg = 'Convolutions on more than 4 dimensions only supported in NHWC.'
    assert self._lhs_spec == self._out_spec == 'NHWC', msg

  def forward(self, x):
    if self._use_bias:
      w, b = self.weights
    else:
      w = self.weights
    x_shape = list(x.shape)
    if len(x_shape) > 4:
      self._check_nhwc()
      new_batch_dim = functools.reduce(operator.mul, x_shape[:-3])
      x = jnp.reshape(x, [new_batch_dim] + x_shape[-3:])
    res = fastmath.conv(
        x, w, self._strides, self._padding, self._dimension_numbers,
        self._one)
    if self._use_bias:
      res = res + b
    if len(x_shape) > 4:
      res = jnp.reshape(res, x_shape[:-3] + list(res.shape[-3:]))
    return res

  def _kernel_shape(self, input_shape):
    """Helper to calculate the kernel shape."""
    kernel_size_iter = iter(self._kernel_size)
    return [self._filters if c == 'O' else
            input_shape[self._lhs_spec.index('C')] if c == 'I' else
            next(kernel_size_iter) for c in self._rhs_spec]

  def init_weights_and_state(self, input_signature):
    input_shape = input_signature.shape
    if len(input_shape) > 4:
      self._check_nhwc()
      new_batch_dim = functools.reduce(operator.mul, input_shape[:-3])
      input_shape = [new_batch_dim] + list(input_shape[-3:])
    kernel_shape = self._kernel_shape(input_shape)
    rng1, rng2 = fastmath.random.split(self.rng, 2)
    w = self._kernel_initializer(kernel_shape, rng1)
    if self._use_bias:
      bias_shape = [self._filters if c == 'C' else 1 for c in self._out_spec]
      bias_shape = tuple(itertools.dropwhile(lambda x: x == 1, bias_shape))
      b = self._bias_initializer(bias_shape, rng2)
      self.weights = (w, b)
    else:
      self.weights = w


class CausalConv(Conv):
  """Causal (masked) convolution for [batch x time x depth] sequences.

  Maintains causality along time axis. Used in language modeling tasks.
  """

  def __init__(self,
               filters,
               kernel_width=3,
               kernel_initializer=None,
               bias_initializer=init.RandomNormalInitializer(1e-6),
               use_bias=True):
    super().__init__(
        filters=filters,
        kernel_size=(kernel_width,),
        strides=None,
        padding='VALID',
        dimension_numbers=('NWC', 'WIO', 'NWC'),
        kernel_initializer=kernel_initializer,
        bias_initializer=bias_initializer,
        use_bias=use_bias)

  def forward(self, x):
    assert self._padding == 'VALID'
    # Left pad with 0s. Applying an unmasked valid convolution on top of this
    # yields a causal convolution.
    # TODO(ddohan): Support strided and dilated convolutions.
    rate = 1
    effective_kernel_size = int((self._kernel_size[0] - 1) * rate + 1)
    pad = effective_kernel_size - 1
    x_leftpad = (
        jnp.pad(x, pad_width=[[0, 0], [pad, 0], [0, 0]], mode='constant'))
    return super().forward(x_leftpad)


def Conv1d(filters, kernel_size, stride=1, padding='VALID',
           kernel_initializer=None,
           bias_initializer=init.RandomNormalInitializer(1e-6),
           use_bias=True):
  return Conv(filters, (kernel_size,), strides=(stride,), padding=padding,
              dimension_numbers=('NWC', 'WIO', 'NWC'),
              kernel_initializer=kernel_initializer,
              bias_initializer=bias_initializer,
              use_bias=use_bias)


def _zero_pad(x, pad, axis):  # pylint: disable = invalid-name
  """Helper for jnp.pad with 0s for single-axis case."""
  pad_widths = [(0, 0)] * len(x.shape)
  pad_widths[axis] = pad  # Padding on axis.
  return jnp.pad(x, pad_widths, mode='constant')


# @assert_shape('bld->bld')
class CausalDepthwiseConv(base.Layer):
  """A causal depthwise convolution layer."""

  def __init__(self,
               kernel_size=3,
               kernel_initializer=init.GlorotUniformInitializer(),
               use_bfloat16=False):
    """Returns a causal depthwise convolution layer."""
    super().__init__(n_in=1, n_out=1)
    self._kernel_size = kernel_size
    self._kernel_initializer = kernel_initializer
    self._use_bfloat16 = use_bfloat16

  def forward(self, x):
    """Executes this layer as part of a forward pass through the model.

    Args:
      x: Tensor of same shape and dtype as the input signature used to
          initialize this layer.

    Returns:
      Tensor of same shape and dtype as the input.
    """
    w = self.weights
    res = x * w[0, :][None, None, :]
    for i in range(1, self._kernel_size):
      x = _zero_pad(x, (1, 0), 1)
      x = x[:, :-1, :]
      res += x * w[i, :][None, None, :]
    return res

  def init_weights_and_state(self, input_signature):
    """Randomly initializes this layer's weights."""
    shape_w = (self._kernel_size, input_signature.shape[-1])
    rng_w, _ = fastmath.random.split(self.rng, 2)
    w = self._kernel_initializer(shape_w, rng_w)
    if self._use_bfloat16:
      w = w.astype(jnp.bfloat16)
    self.weights = w



# coding=utf-8
# Copyright 2021 The Trax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Tests for RL environments created from supervised data-sets."""

from absl.testing import absltest
import numpy as np
from trax.rl.envs import data_envs


class SequenceDataEnvTest(absltest.TestCase):

  def _assert_masks(self, info, control, discount):
    self.assertEqual(info, {'control_mask': control, 'discount_mask': discount})

  def test_copy_task_short_sequence_correct_actions(self):
    """Test sequence data env on the copying task, correct replies.

    With input (x1, x2) this tests for the following sequence of
    (observations, rewards, dones, actions):
    x1                 = env.reset()
    x2,   0.0,  F, _   = env.step(ignored_action)
    eos,  0.0,  F, _   = env.step(ignored_action)
    x1,   0.0,  F, _   = env.step(x1)
    x2,   0.0,  F, _   = env.step(x2)
    eos,  1.0,  T, _   = env.step(eos)
    """
    env = data_envs.SequenceDataEnv(data_envs.copy_stream(2, n=1), 16)
    x1 = env.reset()
    x2, r0, d0, i0 = env.step(0)
    self.assertEqual(r0, 0.0)
    self.assertEqual(d0, False)
    self._assert_masks(i0, control=0, discount=0)
    eos, r1, d1, i1 = env.step(0)
    self.assertEqual(eos, 1)
    self.assertEqual(r1, 0.0)
    self.assertEqual(d1, False)
    self._assert_masks(i1, control=0, discount=0)
    y1, r2, d2, i2 = env.step(x1)
    self.assertEqual(y1, x1)
    self.assertEqual(r2, 0.0)
    self.assertEqual(d2, False)
    self._assert_masks(i2, control=1, discount=0)
    y2, r3, d3, i3 = env.step(x2)
    self.assertEqual(y2, x2)
    self.assertEqual(r3, 0.0)
    self.assertEqual(d3, False)
    self._assert_masks(i3, control=1, discount=0)
    eos2, r4, d4, i4 = env.step(1)
    self.assertEqual(eos2, 1)
    self.assertEqual(r4, 1.0)
    self.assertEqual(d4, True)
    self._assert_masks(i4, control=1, discount=1)

  def test_copy_task_longer_sequnece_mixed_actions(self):
    """Test sequence data env on the copying task, mixed replies.

    With input (x1, x2) and (y1, y2) this tests for the following sequence of
    (observations, rewards, dones, actions):
    x1                 = env.reset()
    x2,   0.0,  F, _   = env.step(ignored_action)
    eos,  0.0,  F, _   = env.step(ignored_action)
    x1,   0.0,  F, _   = env.step(x1)
    x2+1, 0.0,  F, _   = env.step(x2+1)
    y1,   0,5,  F, _   = env.step(eos)
    y2,   0.0,  F, _   = env.step(ignored_action)
    eos,  0.0,  F, _   = env.step(ignored_action)
    y1+1  0.0,  F, _   = env.step(y1+1)
    y2+1, 0.0,  F, _   = env.step(y2+1)
    eos,  0.0,  T, _   = env.step(eos)
    """
    env = data_envs.SequenceDataEnv(data_envs.copy_stream(2, n=2), 16)
    x1 = env.reset()
    x2, _, _, _ = env.step(0)
    eos, _, _, _ = env.step(0)
    _, _, _, _ = env.step(x1)
    _, _, _, _ = env.step(x2 + 1)  # incorrect
    y1, r1, d1, _ = env.step(1)
    self.assertEqual(r1, 0.5)
    self.assertEqual(d1, False)
    y2, _, _, _ = env.step(0)
    eos, _, _, _ = env.step(0)
    _, _, _, _ = env.step(y1 + 1)  # incorrect
    _, _, _, _ = env.step(y2 + 1)  # incorrect
    eos, r2, d2, _ = env.step(1)
    self.assertEqual(eos, 1)
    self.assertEqual(r2, 0.0)
    self.assertEqual(d2, True)

  def test_copy_task_action_observation_space(self):
    """Test that sequence data env returns correct action/observation space."""
    env = data_envs.SequenceDataEnv(data_envs.copy_stream(2, n=1), 16)
    self.assertEqual(env.action_space.n, 16)
    self.assertEqual(env.observation_space.n, 16)

  def test_copy_task_max_length(self):
    """Test that sequence data env respects max_length."""
    env = data_envs.SequenceDataEnv(data_envs.copy_stream(10, n=1), 16,
                                    max_length=2)
    obs = env.reset()
    for _ in range(10):
      obs, reward, done, _ = env.step(0)
      self.assertEqual(reward, 0.0)
      self.assertEqual(done, False)
    self.assertEqual(obs, 1)  # produces EOS
    obs, reward, done, _ = env.step(7)
    self.assertEqual(obs, 7)  # repeats action
    self.assertEqual(reward, 0.0)
    self.assertEqual(done, False)
    obs, reward, done, _ = env.step(8)
    self.assertEqual(obs, 8)  # repeats action
    self.assertEqual(reward, 0.0)
    self.assertEqual(done, False)
    obs, reward, done, _ = env.step(9)
    self.assertEqual(done, True)  # exceeded max_length, stop
    self.assertEqual(obs, 1)      # produce EOS on done
    obs, reward, done, _ = env.step(10)
    self.assertEqual(done, True)  # continue producing done = True
    self.assertEqual(obs, 1)      # continue producing EOS

  def test_number_of_active_masks(self):
    """Test that we have the correct number of control and discount masks."""
    n_input_seqs = 3
    n_output_seqs = 2
    input_len = 4
    output_len = 5

    def data_stream():
      i = 2 * np.ones(input_len)
      o = np.zeros(output_len)
      while True:
        yield (i, o, i, o, i)  # 3 input, 2 output sequences.

    env = data_envs.SequenceDataEnv(data_stream(), 16, max_length=output_len)
    env.reset()

    n_discount = 0
    n_control = 0
    n_steps = 0
    done = False
    while not done:
      (_, _, done, info) = env.step(action=0)
      n_discount += info['discount_mask']
      n_control += info['control_mask']
      n_steps += 1

    # One discount_mask=1 per output sequence.
    self.assertEqual(n_discount, n_output_seqs)
    # One control_mask=1 per output token, including EOS, because it's also
    # controlled by the agent.
    self.assertEqual(n_control, (output_len + 1) * n_output_seqs)
    # One control_mask=0 per input token, excluding EOS, because when the env
    # emits it, control transfers to the agent immediately.
    self.assertEqual(n_steps - n_control, input_len * n_input_seqs)


if __name__ == '__main__':
  absltest.main()



# coding=utf-8
# Copyright 2021 The Trax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Decoding with Trax models."""

import numpy as np
from trax import fastmath
from trax import layers as tl


def autoregressive_sample_stream(model, inputs=None,
                                 batch_size=1, temperature=1.0,
                                 start_id=0, accelerate=True,
                                 eval_mode=False, eval_min_length=1):
  """Yields samples from `model`, in autoregressive language model fashion.

  This function uses `model` to generate outputs one position at a time, with
  access to inputs for the current position and all preceding positions. The
  new output becomes the next position's input, and further calls to
  `autoregressive_sample_stream` repeat the process for successive positions
  indefinitely.

  Inputs and outputs always come in batches, even if size 1. If `inputs` is
  present, it must have shape (`batch_size`, inputs_sequence_length), and each
  output in the stream has shape (`batch_size`, 1).

  Args:
    model: A layer object (subclass of `trax.layers.Layer`) created in
        `'predict'` mode and initialized from trained weights. The model
        must have a structure that allows it to run as an autoregressive
        one-sample-at-a-time predictor (e.g., `trax.models.TransformerLM`),
        except if `eval_mode` is set -- any model can be sampled then,
        but the sampling process may be much slower.
    inputs: Sequence of symbols the model sees as input the first time it
        generates an output. If None, the model generates the first output
        based on just the start symbol.
    batch_size: Number of sequences to generate in parallel as a batch.
    temperature: Parameter that controls the sharpness of the softmax that
        feeds the sampling process. Values range from 0.0 (all probability mass
        goes to one candidate; like an argmax) to positive infinity (all
        candidates have equal probability).
    start_id: Integer representing the start symbol for the autoregressive
        process, or array of shape (`batch_size`, 1) of such integers.
    accelerate: If True, create an accelerated version of `model` and use it
        for generating outputs.
    eval_mode: If True, assume the model is created in `eval` mode and sample
        by collecting all previous outputs and passing the whole tensor.
    eval_min_length: If set, the minimum length to pad to in eval mode.

  Yields:
    Tensor of integers with shape (`batch_size`, 1), representing the batch of
    outputs for the next position in the stream.
  """
  if inputs is not None and inputs.shape[0] != batch_size:
    raise ValueError(f'Inputs batch size ({inputs.shape[0]}) does not match '
                     f'batch_size arg ({batch_size}.')

  fast_model = tl.Accelerate(model) if accelerate else model
  if np.isscalar(start_id):
    start_symbol = np.full((batch_size, 1), start_id, dtype=np.int32)
  else:
    start_symbol = start_id
  if model.n_in == 1 and inputs is not None:
    current_symbols = np.concatenate([start_symbol, inputs], axis=1)
  else:
    current_symbols = start_symbol

  if eval_mode:
    # no start symbol needed in eval mode
    current_symbols = current_symbols[:, 1:]

  while True:
    # Pad inputs to power-of-2 length if needed.
    if eval_mode:
      # one extra symbol as an initial one will be added
      l = max(eval_min_length, current_symbols.shape[1] + 1)
      pad_len = int(2**np.ceil(np.log2(l))) - current_symbols.shape[1]
      unpadded_symbols = current_symbols
      current_symbols = np.pad(
          current_symbols, [[0, 0], [0, pad_len]], mode='constant')
      last_index = -pad_len  # no -1 as the starting one will be added
    else:
      last_index = -1
    # Run the model.
    if model.n_in > 1 and inputs is not None:
      logits = fast_model((inputs, current_symbols))[0]
    else:
      logits = fast_model(current_symbols)
    logits = tl.log_softmax(logits[:, last_index, :])
    sample = tl.logsoftmax_sample(logits, temperature=temperature)
    yield sample
    if eval_mode:
      current_symbols = np.concatenate(
          [unpadded_symbols, sample[:, None]], axis=1)
    else:
      # NOTE: Because the model is autoregressive and in 'predict' mode, its
      # history is cached in the model state and the next input is the single
      # symbol just sampled.
      current_symbols = sample[:, None]


def autoregressive_sample(model, inputs=None,
                          batch_size=1, temperature=1.0,
                          start_id=0, eos_id=1, max_length=100,
                          accelerate=True, eval_mode=False, eval_min_length=1):
  """Returns a batch of sequences created by autoregressive sampling.

  This function uses `model` to generate outputs one position at a time, with
  access to inputs for the current position and all preceding positions. The
  new output becomes the next position's input, and this loop repeats until
  either the model outputs the `eos_id` value or the output sequence reaches
  `max_length` items.

  Args:
    model: A layer object (subclass of `trax.layers.Layer`) created in
        `'predict'` mode and initialized from trained weights. The model
        must have a structure that allows it to run as autoregressive
        one-sample-at-a-time predictor (e.g., `trax.models.TransformerLM`),
        except if `eval_mode` is set -- any model can be sampled then,
        but the sampling process may be much slower.
    inputs: Sequence of symbols the model sees as input the first time it
        generates an output. If None, the model must generate the first output
        with no input to guide it.
    batch_size: Number of sequences to generate in parallel as a batch.
    temperature: Parameter that controls the sharpness of the softmax that
        feeds the sampling process. Values range from 0.0 (all probability mass
        goes to one candidate; like an argmax) to positive infinity (all
        candidates have equal probability).
    start_id: The start symbol (ID/integer) for the autoregressive process,
        or array of shape (`batch_size`, 1) of such integers.
    eos_id: The end-of-sequence symbol (ID/integer) for the autoregressive
        process.
    max_length: Maximum length for generated sequences.
    accelerate: If True, create an accelerated version of `model` and use it
        for generating outputs.
    eval_mode: If True, assume the model is created in `eval` mode and sample
        by collecting all previous outputs and passing the whole tensor.
    eval_min_length: If set, the minimum length to pad to in eval mode.

  Returns:
    Tensor of integers with shape (`batch_size`, output_length) representing
    a batch of output sequences. output_length is the maximum length of the
    output sequences, where each sequence can be no longer than `max_length`.
  """
  result = []
  eos_seen = []
  counter = 0
  for sample in autoregressive_sample_stream(
      model, inputs, batch_size=batch_size, temperature=temperature,
      start_id=start_id, accelerate=accelerate, eval_mode=eval_mode,
      eval_min_length=eval_min_length):
    sample = sample[:, None]
    result.append(sample)
    counter += 1
    if counter >= max_length:
      return np.concatenate(result, axis=1)
    # Check at which batch positions have we already encountered EOS.
    for j in range(batch_size):
      if int(sample[j, 0]) == eos_id:
        eos_seen.append(j)
    # If EOS has been seen on all positions, stop.
    if all([j in eos_seen for j in range(batch_size)]):
      return np.concatenate(result, axis=1)
  return np.concatenate(result, axis=1)


def beam_search(model, inputs=None, batch_size=1, n_beams=2, start_id=0,
                eos_id=1, max_length=100, length_penalty=1.0, accelerate=True):
  """Returns a batch of n_beams-sequences created by beam search.

  This function uses `model` to generate outputs one position at a time, with
  access to inputs for the current position and all preceding positions. The
  new output becomes the next position's input, and this loop repeats until
  either the model outputs the `eos_id` value or the output sequence reaches
  `max_length` items -- but keeping n_beams top beams.

  Args:
    model: A layer object (subclass of `trax.layers.Layer`) created in
        `'predict'` mode and initialized from trained weights. The model
        must have a structure that allows it to run as autoregressive
        one-sample-at-a-time predictor (e.g., `trax.models.TransformerLM`).
    inputs: Sequence of symbols the model sees as input the first time it
        generates an output. If None, the model must generate the first output
        with no input to guide it.
    batch_size: Number of sequences to generate in parallel as a batch.
    n_beams: How many beams to consider at the same time.
    start_id: The start symbol (ID/integer) for the autoregressive process,
        or array of shape (`batch_size`, 1) of such integers.
    eos_id: The end-of-sequence symbol (ID/integer) for the autoregressive
        process.
    max_length: Maximum length for generated sequences.
    length_penalty: Factor alpha in calculating the length penalty for beams.
    accelerate: If True, create an accelerated version of `model` and use it
        for generating outputs.

  Returns:
    Tensor of integers with shape (`batch_size`, n_beams, output_length) with
    a batch of output sequences. output_length is the maximum length of the
    output sequences, where each sequence can be no longer than `max_length`.
  """
  del eos_id, length_penalty  # TODO(lukaszkaiser): add length penalty, eos
  assert batch_size == 1, 'Batch size > 1 not supported yet'
  if inputs is not None and inputs.shape[0] != batch_size:
    raise ValueError(f'Inputs batch size ({inputs.shape[0]}) does not match '
                     f'batch_size arg ({batch_size}.')

  fast_model = tl.Accelerate(model) if accelerate else model
  if np.isscalar(start_id):
    start_symbol = np.full((batch_size, 1), start_id, dtype=np.int32)
  else:
    start_symbol = start_id
  if model.n_in == 1 and inputs is not None:
    current_symbols = np.concatenate([start_symbol, inputs], axis=1)
  else:
    current_symbols = start_symbol

  beams = [current_symbols for _ in range(n_beams)]
  results = [([], 0.0) for _ in range(n_beams)]
  states = [fast_model.state for _ in range(n_beams)]
  top_k = [None] * n_beams
  counter = 0
  while counter < max_length:
    counter += 1
    # Run the model on all beams, collect states and top_k for each beam.
    for beam_id in range(n_beams if counter > 1 else 1):
      fast_model.state = states[beam_id]
      if model.n_in > 1 and inputs is not None:
        logits = fast_model((inputs, beams[beam_id]))[0]
      else:
        logits = fast_model(beams[beam_id])
      logits = tl.log_softmax(logits[:, -1, :])
      states[beam_id] = fast_model.state
      top_k[beam_id] = fastmath.top_k(logits, k=n_beams)

    # Select new beams.
    cur_values = []  # will hold triples (sum-of-logprobs, beam-id, symbol)
    for beam_id in range(n_beams if counter > 1 else 1):
      for k in range(n_beams):
        values, symbols = top_k[beam_id]
        value, symbol = values[:, k], symbols[:, k]
        cur_values.append((results[beam_id][1] + value, beam_id, symbol))
    cur_values.sort(key=lambda x: -x[0][0])  # x[0][0] as batch_size=1
    # Collect top beams to the new states and results.
    new_results, new_states, new_beams = [], [], []
    for (value, beam_id, symbol) in cur_values[:n_beams]:
      new_results.append((results[beam_id][0] + [symbol], value))
      new_states.append(states[beam_id])  # copy?
      new_beams.append(symbol[:, None])
    results, states, beams = new_results, new_states, new_beams

  return [(np.stack(r, axis=-1), v) for (r, v) in results]



# coding=utf-8
# Copyright 2021 The Trax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Adam optimizer class."""

from trax.fastmath import numpy as jnp
from trax.optimizers import base as opt_base


# pylint: disable=line-too-long
class Adam(opt_base.Optimizer):
  r"""Adam optimizer; described in https://arxiv.org/abs/1412.6980.

  The update rule for time step :math:`t`, given gradients :math:`g_t` and
  "Stepsize" :math:`\alpha`, is:

  .. math::
      \hat{m}_t &\leftarrow \big(\beta_1 \cdot m_{t-1} + (1 - \beta_1) \cdot g_t\big)\ /\ (1 - \beta_1^t) \\
      \hat{v}_t &\leftarrow \big(\beta_2 \cdot m_{t-1} + (1 - \beta_2) \cdot g_t^2\big)\ /\ (1 - \beta_2^t) \\
      \theta_t  &\leftarrow \theta_{t-1} -\ \alpha \cdot \hat{m}_t / \big(\sqrt{\hat{v}_t} + \epsilon\big)

  """
  # pylint: enable=line-too-long

  def __init__(self, learning_rate=0.0001, weight_decay_rate=1e-5,  # pylint: disable=useless-super-delegation
               b1=0.9, b2=0.999, eps=1e-5, clip_grad_norm=None):
    r"""Creates an Adam optimizer.

    Args:
      learning_rate: Initial (unadapted) learning rate :math:`\alpha`; original
          paper calls this `Stepsize` and suggests .001 as a generally good
          value.
      weight_decay_rate: Fraction of prior weight values to subtract on each
          step; equivalent to multiplying each weight element by
          `1 - weight_decay_rate`. (This is not part of the core Adam
          algorithm.)
      b1: Exponential decay rate :math:`\beta_1` for first moment estimates.
      b2: Exponential decay rate :math:`\beta_2` for second moment estimates.
      eps: Small positive constant :math:`\epsilon` for numerical stability.
      clip_grad_norm: Threshold value above which gradient clipping occurs.
          (This is not part of the core Adam algorithm.)
    """
    super().__init__(
        learning_rate=learning_rate,
        weight_decay_rate=weight_decay_rate,
        b1=b1,
        b2=b2,
        eps=eps,
        clip_grad_norm=clip_grad_norm
    )

  def init(self, weights):
    m = jnp.zeros_like(weights)
    v = jnp.zeros_like(weights)
    return m, v

  def update(self, step, grads, weights, slots, opt_params):
    m, v = slots
    learning_rate = opt_params['learning_rate']
    weight_decay_rate = opt_params['weight_decay_rate']
    b1 = opt_params['b1']
    b2 = opt_params['b2']
    eps = opt_params['eps']
    m = (1 - b1) * grads + b1 * m  # First  moment estimate.
    v = (1 - b2) * (grads ** 2) + b2 * v  # Second moment estimate.
    mhat = m / (1 - b1 ** (step + 1))  # Bias correction.
    vhat = v / (1 - b2 ** (step + 1))
    new_weights = ((1 - weight_decay_rate) * weights - (
        learning_rate * mhat / (jnp.sqrt(vhat) + eps))).astype(weights.dtype)
    return new_weights, (m, v)



# coding=utf-8
# Copyright 2021 The Trax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Resampling module."""

from trax.fastmath import numpy as jnp
from trax.layers import combinators as cb
from trax.layers import core
from trax.layers.normalization import LayerNorm
from trax.layers.pooling import AvgPool
from trax.layers.research.rel_attention import RelativeAttentionLMLayer


def AveragePooling(shorten_factor, *args, **kwargs):
  del args, kwargs

  return AvgPool(pool_size=(shorten_factor,), strides=(shorten_factor,))


def LinearPooling(shorten_factor, d_model, *args, dropout=0.0, mode='train',
                  **kwargs):
  del args, kwargs

  return cb.Serial(
      core.Fn(
          'Shorten',
          lambda x: jnp.reshape(  # pylint: disable=g-long-lambda
              # Shorten -- move to depth.  # pylint: disable=g-long-lambda
              x, (x.shape[0], x.shape[1] // shorten_factor, -1)),
          n_out=1),
      core.Dense(d_model),
      core.Dropout(rate=dropout, mode=mode)
  )


def LinearUpsampling(shorten_factor, d_model, *args, dropout=0.0, mode='train',
                     **kwargs):
  del args, kwargs

  return cb.Serial(
      core.Dense(shorten_factor * d_model),
      core.Dropout(rate=dropout, mode=mode),
      core.Fn(
          'ProlongBack',
          lambda x: jnp.reshape(  # pylint: disable=g-long-lambda
              # Prolong back.  # pylint: disable=g-long-lambda
              x, (x.shape[0], x.shape[1] * shorten_factor, -1)),
          n_out=1)
  )


def NaiveUpsampling(shorten_factor, d_model, *args, **kwargs):  # pylint: disable = unused-argument
  return core.Fn('Repeat', lambda x: jnp.repeat(x, shorten_factor, axis=1))


def NoUpsampling(shorten_factor, d_model, *args, **kwargs):
  del d_model, args, kwargs

  return core.Fn('ReturnZero', lambda x: jnp.zeros(  # pylint: disable=g-long-lambda
      (x.shape[0], x.shape[1] * shorten_factor, x.shape[2]), dtype=x.dtype))


def FeedForwardBlock(d_model,
                     d_ff,
                     dropout,
                     dropout_shared_axes,
                     mode,
                     activation):
  # We copy the ff block function because we cannot import it from models
  return [
      core.Dense(d_ff),
      activation(),
      core.Dropout(rate=dropout, shared_axes=dropout_shared_axes,
                   mode=mode),
      core.Dense(d_model),
  ]


def AttentionResampling(shorten_factor, d_model, is_upsampling, d_ff, n_heads,
                        dropout, dropout_shared_axes, mode, ff_activation,
                        context_bias_layer, location_bias_layer, total_pooling,
                        resampling_fn):
  """Attention resampling."""

  attention = RelativeAttentionLMLayer(
      d_model, context_bias_layer, location_bias_layer,
      total_pooling, n_heads=n_heads, dropout=dropout,
      mode=mode)

  feed_forward = FeedForwardBlock(
      d_model, d_ff, dropout, dropout_shared_axes, mode, ff_activation)

  resampling = resampling_fn(shorten_factor, d_model,
                             mode=mode)

  def _Dropout():
    return core.Dropout(rate=dropout, shared_axes=dropout_shared_axes,
                        mode=mode)

  return [
      LayerNorm(),               # h
      cb.Branch(cb.Serial(
          resampling,
          LayerNorm(),
      ), None),                  # h', h
      cb.Serial(  # pylint: disable=g-long-ternary
          cb.Select([0, 2, 1, 2]),
          cb.Add(),
      ) if is_upsampling else [],
      cb.Residual(
          cb.Select([0, 1, 1]),  # h', h, h
          attention,
          _Dropout(),
      ),
      cb.Residual(
          LayerNorm(),
          feed_forward,
          _Dropout(),
      ),
  ]



# coding=utf-8
# Copyright 2021 The Trax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Trax base optimizer class."""

from trax import fastmath
from trax.fastmath import numpy as jnp


class Optimizer:
  """Base class for optimizers that work hand in hand with Trax layers.

  To define an optimizer subclass, specify its behavior with respect to a
  single node in the network (e.g., a single dense layer):

    - `init`: how to create/initialize optimizer-internal parameters ("slots"),
        as a function of the node's weights.
    - `update`: how to use gradient information to update node weights and
        optimizer slots.

  The Trax runtime combines these node-local computations into layer weight
  updates and optimizer slot updates for the whole tree of layers in the model.
  """

  def __init__(self, learning_rate=0.01, clip_grad_norm=None,
               **init_opt_params):
    """Sets initial hyperparameter values for this optimizer.

    Takes optimizer hyperparameters as keyword arguments. These values can
    change over time (training steps), e.g., for learning rate schedules.

    To expose subclass hyperparameters for gin configuration, override this
    constructor and use explicitly named keyword arguments. See
    `momentum.Momentum.__init__` for one such example.

    Args:
      learning_rate: Learning rate for the optimizer. This can change during
          training by means of a training rate schedule.
      clip_grad_norm: If specified, this scalar value is used to limit gradient
          size -- all gradient elements in a training step are treated as if
          they belonged to a single vector and then scaled back if needed so
          that such a vector's L2 norm does not exceed `clip_grad_norm`. If
          None, no clipping happens.
      **init_opt_params: Initial values of any additional optimizer parameters.
    """
    init_opt_params['learning_rate'] = learning_rate
    self._init_opt_params = {
        name: jnp.array(value) for (name, value) in init_opt_params.items()
    }
    self._slots = None
    # Gradient clipping happens with respect to the norm of the whole gradient
    # tree, so it is not passed to single-slot updates, but done in this class
    # for the whole gradient tree.
    self._clip_grad_norm = clip_grad_norm

  def init(self, weights):
    """Creates optimizer slots that fit the given weights.

    Args:
      weights: Trainable weights for one layer. Optimizer slots typically match
          the data shape and type of the given layer weights.
    """
    raise NotImplementedError

  def update(self, step, grads, weights, slots, opt_params):
    """Computes updated layer weights and optimizer slots for one training step.

    Args:
      step: Training step number.
      grads: Gradient values for this node (from back-propagation during a
          training step).
      weights: Current weight values for this node (i.e., layer weights).
      slots: Current slot values for this node.
      opt_params: Optimizer hyperparameters (e.g. learning rate, momentum),
          same across all nodes in the model.

    Returns:
      Tuple of (new_weights, new_slots), which the Trax runtime will use to
      update the model and optimizer within each training step.
    """
    raise NotImplementedError

  @property
  def slots(self):
    return self._slots

  @slots.setter
  def slots(self, slots):
    self._slots = slots

  @property
  def opt_params(self):
    return self._init_opt_params

  @opt_params.setter
  def opt_params(self, opt_params):
    self._init_opt_params = opt_params

  def tree_init(self, weight_tree):
    """Assembles node-local initializations into full-tree initialization.

    Args:
      weight_tree: Weights for an entire model, in a tree that matches the
          model's layer structure.

    Returns:
      Tuple `(slots, opt_params)`, where `slots` are the initialized optimizer
      slot values and `opt_params` are optimizer hyperparameters (e.g.,
      learning rate, momentum).
    """
    self._slots = tuple(self.init(weight)
                        for weight in fastmath.tree_flatten(weight_tree))
    return (self._slots, self._init_opt_params)

  def tree_update(self, step, grad_tree, weight_tree, slots, opt_params,
                  store_slots=True):
    """Assembles node-local weight and slot updates for the full layer tree.

    Args:
      step: Current step number in the training process.
      grad_tree: Gradients for the entire model, in a tree that matches the
          model's layer structure.
      weight_tree: Current weights for the entire model, in a tree that matches
          the model's layer structure.
      slots: Optimizer slots.
      opt_params: Optimizer hyperparameters (e.g. learning rate, momentum).
      store_slots: Boolean; if True, stores resulting slots in this object;
        when set to False, this becomes a pure function.

    Returns:
      Tuple `(weights, slots)`, where `weights` are the optimizer-updated
      weights for the whole model (in a tree matching the model's layer
      structure) and `slots` are the updated optimizer slot values.
    """
    grads_flat = fastmath.tree_flatten(grad_tree)
    grads_norm = self._l2_norm(grads_flat)
    if self._clip_grad_norm is not None:
      max_norm = self._clip_grad_norm
      grads_flat = [jnp.where(grads_norm < max_norm,  # pylint: disable=g-complex-comprehension
                              g,
                              g * (max_norm / grads_norm))
                    for g in grads_flat]
    weights_flat = fastmath.tree_flatten(weight_tree)
    weights_norm = self._l2_norm(weights_flat)
    updated_pairs = [
        self._update_and_check(step, grad, weight, slot, opt_params)
        for (grad, weight, slot) in zip(grads_flat, weights_flat, slots)
    ]
    new_weights_flat, slots = zip(*updated_pairs)
    new_weights, _ = fastmath.tree_unflatten(new_weights_flat, weight_tree)
    metrics = {'gradients_l2': grads_norm, 'weights_l2': weights_norm}
    slots = tuple(slots)
    if store_slots:
      self.slots = slots
    return new_weights, slots, metrics

  def _l2_norm(self, flat_list):
    """Returns an L2-like norm of all elements of all tensors in `flat_list`.

    Args:
      flat_list: Collection of tensors as a flat list (rather than, e.g., a
          tree).

    Returns:
      A scalar value computed as if all the tensors in `flat_list` were joined
      and flattened into a single vector, and then the L2 norm of that vector
      was calculated.
    """
    if fastmath.is_backend(fastmath.Backend.JAX):
      norm = jnp.sqrt(sum(jnp.vdot(x, x) for x in flat_list))
    else:  # TODO(lukaszkaiser): add vdot to TF-numpy
      norm = jnp.sqrt(sum(jnp.sum(x*x) for x in flat_list))
    return norm

  def _update_and_check(self, step, grads, weights, slots, opt_params):
    """Updates a single weight array and checks types."""
    new_weights, new_slots = self.update(
        step, grads, weights, slots, opt_params)
    if isinstance(weights, jnp.ndarray):
      if not isinstance(new_weights, jnp.ndarray):
        raise ValueError(
            f'New weight values should be of type jnp.ndarray or a subclass; '
            f'instead got {type(new_weights)}.')
      if new_weights.dtype != weights.dtype:
        raise ValueError(
            f'New weight values dtype ({new_weights.dtype}) does not match '
            f'the old one ({weights.dtype}).')
    return new_weights, new_slots


class SGD(Optimizer):
  """Stochastic gradient descent (SGD) optimizer."""

  def init(self, weights):
    return None

  def update(self, step, grads, weights, slots, opt_params):
    del step, slots
    lr = opt_params['learning_rate']
    new_weights = weights - (lr * grads).astype(weights.dtype)
    return new_weights, None


# Utilities.


def l2_norm(tree):
  """Returns an L2 norm computed over all elements of all tensors in `tree`.

  Args:
    tree: Tree-structured collection of tensors, e.g., model weights matching
        the model's layer structure.

  Returns:
    A scalar value computed as if all the tensors in `tree` were combined
    and flattened into a single vector, and then the L2 norm of that vector
    was calculated.
  """
  leaves = fastmath.tree_flatten(tree)
  return jnp.sqrt(sum(jnp.vdot(x, x) for x in leaves))


def clip_grads(grad_tree, max_norm):
  """Proportionally reduces each gradient value to respect an aggregate limit.

  Args:
    grad_tree: Gradient values structured as a tree of tensors matching the
        model's layer structure.
    max_norm: The aggregate limit on gradient values. All gradient elements in
        `grad_tree` are treated as if they belonged to a single vector and
        that vector is shortened if needed so that its L2 norm does not exceed
        `clip_grad_norm`.

  Returns:
    A new tree of tensors matching the structure of `grad_tree`, but with
    element values proportionally rescaled as needed to respect the `max_norm`
    limit.
  """
  norm = l2_norm(grad_tree)
  normalize = lambda g: jnp.where(norm < max_norm, g, g * (max_norm / norm))
  return fastmath.nested_map(grad_tree, normalize)
